"""Default configs for image classification"""
# pylint: disable=bad-whitespace,missing-class-docstring
from typing import Union, Tuple
from autocfg import dataclass, field


@dataclass
class ImageClassification:
    model : str = 'resnet50_v1'
    use_pretrained : bool = True
    use_gn : bool = False
    batch_norm : bool = False
    use_se : bool = False
    last_gamma : bool = False

@dataclass
class TrainCfg:
    pretrained_base : bool = True  # whether load the imagenet pre-trained base
    batch_size : int = 128
    epochs : int = 10
    lr : float = 0.1  # learning rate
    lr_decay : float = 0.1  # decay rate of learning rate.
    lr_decay_period : int = 0
    lr_decay_epoch : str = '40, 60'  # epochs at which learning rate decays
    lr_mode : str = 'step'  # learning rate scheduler mode. options are step, poly and cosine
    warmup_lr : float = 0.0  # starting warmup learning rate.
    warmup_epochs : int = 0  # number of warmup epochs
    num_training_samples : int = 1281167
    num_workers : int = 4
    wd : float = 0.0001
    momentum : float = 0.9
    teacher : Union[None, str] = None
    hard_weight : float = 0.5
    dtype : str = 'float32'
    input_size : int = 224
    crop_ratio : float = 0.875
    use_rec : bool = False
    rec_train : str = '~/.mxnet/datasets/imagenet/rec/train.rec'
    rec_train_idx : str = '~/.mxnet/datasets/imagenet/rec/train.idx'
    rec_val : str = '~/.mxnet/datasets/imagenet/rec/val.rec'
    rec_val_idx : str = '~/.mxnet/datasets/imagenet/rec/val.idx'
    data_dir : str = '~/.mxnet/datasets/imagenet'
    mixup : bool = False
    no_wd : bool = False
    label_smoothing : bool = False
    temperature : Union[int, float] = 20
    hard_weight : float = 0.5
    resume_epoch : int = 0
    mixup_alpha : float = 0.2
    mixup_off_epoch : int = 0
    log_interval : int = 50
    mode : str = ''
    start_epoch : int = 0
    transfer_lr_mult : float = 0.01  # reduce the backbone lr_mult to avoid quickly destroying the features
    output_lr_mult : float = 0.1  # the learning rate multiplier for last fc layer if trained with transfer learning
    early_stop_patience : int = -1  # epochs with no improvement after which train is early stopped, negative: disabled
    early_stop_min_delta : float = 0.001  # ignore changes less than min_delta for metrics
    # the baseline value for metric, training won't stop if not reaching baseline
    early_stop_baseline : Union[float, int] = 0.0
    early_stop_max_value : Union[float, int] = 1.0  # early stop if reaching max value instantly

@dataclass
class ValidCfg:
    batch_size : int = 128
    num_workers : int = 4

@dataclass
class ImageClassificationCfg:
    img_cls : ImageClassification = field(default_factory=ImageClassification)
    train : TrainCfg = field(default_factory=TrainCfg)
    valid : ValidCfg = field(default_factory=ValidCfg)
    gpus : Union[Tuple, list] = (0, )  # gpu individual ids, not necessarily consecutive



"""Custom and derived deformable convolution"""
# pylint: disable=missing-function-docstring,bad-continuation
import torch
from torch import nn
from torch.nn.modules.utils import _pair
from torchvision.ops import deform_conv2d

__all__ = ['DeformConvWithChangeableStride']


class _NewEmptyTensorOp(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, new_shape):
        ctx.shape = x.shape
        return x.new_empty(new_shape)

    @staticmethod
    def backward(ctx, grad):
        shape = ctx.shape
        return _NewEmptyTensorOp.apply(grad, shape), None


class DeformConvWithChangeableStride(nn.Module):
    """A deformable conv layer with dynamic stride.

    See "https://github.com/pytorch/vision/blob/master/torchvision/ops/deform_conv.py" for usage.
    There are currently some limitations, e.g. deformable_groups must be 1.
    """
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        padding=0,
        dilation=1,
        groups=1,
        deformable_groups=1,
        bias=False,
        norm=None,
        activation=None,
    ):
        super(DeformConvWithChangeableStride, self).__init__()

        assert in_channels % groups == 0, "in_channels {} cannot be divisible by groups {}".format(
            in_channels, groups
        )
        assert (
            out_channels % groups == 0
        ), "out_channels {} cannot be divisible by groups {}".format(out_channels, groups)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _pair(kernel_size)
        self.padding = _pair(padding)
        self.dilation = _pair(dilation)
        self.groups = groups
        self.deformable_groups = deformable_groups
        # TODO(zhreshold): torchvision version of deform_conv limitation
        assert self.deformable_groups == 1, "deformable_groups can only be 1 at this time"
        self.norm = norm
        self.activation = activation

        self.weight = nn.Parameter(
            torch.Tensor(out_channels, in_channels // self.groups, *self.kernel_size)
        )
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_channels))
        else:
            self.bias = None

        nn.init.kaiming_uniform_(self.weight, nonlinearity="relu")
        nn.init.constant_(self.bias, 0.)

    def forward(self, x, offset, stride):
        if x.numel() == 0:
            # When input is empty, we want to return a empty tensor with "correct" shape,
            # So that the following operations will not panic
            # if they check for the shape of the tensor.
            # This computes the height and width of the output tensor
            output_shape = [
                (i + 2 * p - (di * (k - 1) + 1)) // s + 1
                for i, p, di, k, s in zip(
                    x.shape[-2:], self.padding, self.dilation, self.kernel_size, self.stride
                )
            ]
            output_shape = [x.shape[0], self.weight.shape[0]] + output_shape
            return _NewEmptyTensorOp.apply(x, output_shape)

        stride = _pair(stride)
        x = deform_conv2d(
            x,
            offset,
            self.weight,
            self.bias,
            stride,
            self.padding,
            self.dilation,
        )
        if self.norm is not None:
            x = self.norm(x)
        if self.activation is not None:
            x = self.activation(x)
        return x



"""Tools for KITTI Dataset
Code partially borrowed from
https://github.com/nianticlabs/monodepth2/blob/master/kitti_utils.py
"""
from __future__ import absolute_import, division, print_function

import os
from collections import Counter
import numpy as np

import mxnet as mx


def load_velodyne_points(filename):
    """Load 3D point cloud from KITTI file format
    (adapted from https://github.com/hunse/kitti)
    """
    points = np.fromfile(filename, dtype=np.float32).reshape(-1, 4)
    points[:, 3] = 1.0  # homogeneous
    return points


def read_calib_file(path):
    """Read KITTI calibration file
    (from https://github.com/hunse/kitti)
    """
    float_chars = set("0123456789.e+- ")
    data = {}
    with open(path, 'r') as f:
        for line in f.readlines():
            key, value = line.split(':', 1)
            value = value.strip()
            data[key] = value
            if float_chars.issuperset(value):
                # try to cast to float array
                try:
                    data[key] = np.array(list(map(float, value.split(' '))))
                except ValueError:
                    # casting error: data[key] already eq. value, so pass
                    pass

    return data


def sub2ind(matrixSize, rowSub, colSub):
    """Convert row, col matrix subscripts to linear indices
    """
    # pylint: disable=unused-variable
    m, n = matrixSize
    return rowSub * (n - 1) + colSub - 1


def generate_depth_map(calib_dir, velo_filename, cam=2, vel_depth=False):
    """Generate a depth map from velodyne data
    """
    # pylint: disable=unsupported-assignment-operation
    # load calibration files
    cam2cam = read_calib_file(os.path.join(calib_dir, 'calib_cam_to_cam.txt'))
    velo2cam = read_calib_file(os.path.join(calib_dir, 'calib_velo_to_cam.txt'))
    velo2cam = np.hstack((velo2cam['R'].reshape(3, 3), velo2cam['T'][..., np.newaxis]))
    velo2cam = np.vstack((velo2cam, np.array([0, 0, 0, 1.0])))

    # get image shape
    im_shape = cam2cam["S_rect_02"][::-1].astype(np.int32)

    # compute projection matrix velodyne->image plane
    R_cam2rect = np.eye(4)
    R_cam2rect[:3, :3] = cam2cam['R_rect_00'].reshape(3, 3)
    P_rect = cam2cam['P_rect_0' + str(cam)].reshape(3, 4)
    P_velo2im = np.dot(np.dot(P_rect, R_cam2rect), velo2cam)

    # load velodyne points and remove all behind image plane (approximation)
    # each row of the velodyne data is forward, left, up, reflectance
    velo = load_velodyne_points(velo_filename)
    velo = velo[velo[:, 0] >= 0, :]

    # project the points to the camera
    velo_pts_im = np.dot(P_velo2im, velo.T).T
    velo_pts_im[:, :2] = velo_pts_im[:, :2] / velo_pts_im[:, 2][..., np.newaxis]

    if vel_depth:
        velo_pts_im[:, 2] = velo[:, 0]

    # check if in bounds
    # use minus 1 to get the exact same value as KITTI matlab code
    velo_pts_im[:, 0] = np.round(velo_pts_im[:, 0]) - 1
    velo_pts_im[:, 1] = np.round(velo_pts_im[:, 1]) - 1
    val_inds = (velo_pts_im[:, 0] >= 0) & (velo_pts_im[:, 1] >= 0)
    val_inds = val_inds & (velo_pts_im[:, 0] < im_shape[1]) & (velo_pts_im[:, 1] < im_shape[0])
    velo_pts_im = velo_pts_im[val_inds, :]

    # project to image
    depth = np.zeros((im_shape[:2]))
    depth[velo_pts_im[:, 1].astype(np.int), velo_pts_im[:, 0].astype(np.int)] = velo_pts_im[:, 2]

    # find the duplicate points and choose the closest depth
    inds = sub2ind(depth.shape, velo_pts_im[:, 1], velo_pts_im[:, 0])
    dupe_inds = [item for item, count in Counter(inds).items() if count > 1]
    for dd in dupe_inds:
        pts = np.where(inds == dd)[0]
        x_loc = int(velo_pts_im[pts[0], 0])
        y_loc = int(velo_pts_im[pts[0], 1])
        depth[y_loc, x_loc] = velo_pts_im[pts, 2].min()
    depth[depth < 0] = 0

    return depth


def dict_batchify_fn(data):
    """dict batchify function"""
    import mxnet.numpy as _mx_np
    from mxnet.util import is_np_array

    if isinstance(data[0], mx.nd.NDArray):
        return _mx_np.stack(data) if is_np_array() else mx.nd.stack(*data)
    elif isinstance(data[0], dict):
        elem = data[0]
        return {key: dict_batchify_fn([d[key] for d in data]) for key in elem}

    raise RuntimeError('unknown datatype')


def readlines(filename):
    """Read all the lines in a text file and return as a list
    """
    with open(filename, 'r') as f:
        lines = f.read().splitlines()
    return lines



"""
Utility functions for SMOT: Single-Shot Multi Object Tracking
https://arxiv.org/abs/2010.16031
"""
import time
import logging
from contextlib import contextmanager
import mxnet as mx
import numpy as np

def timeit(method):
    """
    The timing decorator to wrap the functions
    """
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        logging.info("{} runtime: {:.04f} msec".format(method.__name__, (te - ts) * 1000))
        return result
    return timed


@contextmanager
def timeit_context(name):
    startTime = time.time()
    yield
    elapsedTime = time.time() - startTime
    logging.info('[{}] runtime {:.03f} msec'.format(name, elapsedTime * 1000))


def mxnet_frame_preprocessing(image, base_size, ratio, mean, std, ctx):
    """
    Parameters
    ----------
    image
    base_size
    ratio: aspect ratio
    mean
    std
    ctx

    Returns
    -------

    """
    image_tensor = mx.nd.array(image, ctx=ctx, dtype=np.uint8)
    float_image = (image_tensor.astype(np.float32) / 255. - mean.reshape((1, 1, 3))) / std.reshape((1, 1, 3))

    trans_image = float_image.transpose((2, 0, 1))
    out_w, out_h = int(base_size * ratio), int(base_size)
    in_h, in_w, _ = image.shape
    pad_w, pad_h = int(max(in_w, in_h * ratio)), int(max(in_h, in_w / ratio))
    pb_w, pb_h = (pad_w - in_w) // 2, (pad_h - in_h) // 2

    # do padding
    padded_image = mx.nd.zeros((1, 3, pad_h, pad_w), ctx=ctx)
    padded_image[0, :, pb_h: pb_h + in_h, pb_w: pb_w + in_w] = trans_image

    # do resizing
    resize_image = mx.nd.contrib.BilinearResize2D(padded_image,
                                                  height=out_h, width=out_w)

    return resize_image, pad_w, pad_h, (pb_w, pb_h, pad_w, pad_h)


def remap_bboxes(bboxes, padded_w, padded_h, expand, data_shape, ratio):
    """
    Remap bboxes in (x0, y0, x1, y1) format into the input image space
    Parameters
    ----------
    bboxes
    padded_w
    padded_h
    expand

    Returns
    -------

    """
    bboxes[:, 0] *= padded_w / (data_shape * ratio)
    bboxes[:, 1] *= padded_h / data_shape
    bboxes[:, 2] *= padded_w / (data_shape * ratio)
    bboxes[:, 3] *= padded_h / data_shape

    bboxes[:, 0] -= expand[0]
    bboxes[:, 1] -= expand[1]
    bboxes[:, 2] -= expand[0]
    bboxes[:, 3] -= expand[1]

    return bboxes


class TrackState:
    """
    States of the track.
    The track follows the simple state machine as below:

    Active: time_since_update always set to 1
        1. If confidence < keep_alive_threshold, goto Missing
        2. If the track is suppressed in track NMS, goto Missing
    Missing: every timestep the missing track increment time_since_update by one
        1. If the track is updated again, goto Active
        2. If time_since_update > max_missing, goto Deleted
    Deleted: This is an absorbing state
    """
    Active = 1
    Missing = 2
    Deleted = 3


class Track:
    """
    This class represents a track/tracklet used in the SMOT Tracker
    It has the following properties

    *******************************************************
    mean: 4-tuple representing the (x0, y0, x1, y1) as the current state (location) of the tracked object
    track_id: the numerical id of the track
    age: the number of timesteps since its first occurrence
    time_since_update: number of time-steps since the last update of the its location
    state: the state of the track, can be one in `TrackState`
    confidence_score: tracking_confidence at the current timestep

    source: a tuple of (anchor_indices, anchor_weights)
    attributes: np.ndarray of additional attributes of the object
    *******************************************************

    It also has these configs
    keep_alive_thresh: the minimal tracking/detection confidence to keep the track in Active state
    max_missing: the maximal timesteps we will keep searching for this track when missing before we mark it as deleted
    *******************************************************

    """

    def __init__(self, mean, track_id, source, keep_alive_thresh=0.1, max_missing=30,
                 attributes=None, class_id=0, linked_id=None):
        self.mean = mean
        self.track_id = track_id
        self.hits = 1
        self.age = 1
        self.time_since_update = 0

        self.state = TrackState.Active
        self.confidence_score = 1.

        self.matched = False
        self.match_score = 999
        self.current_detection_index = None
        self.max_missing = max_missing

        self.next_frame_anchor_id = -1
        self.keep_alive_thresh = keep_alive_thresh
        self.attributes = attributes
        self.source = source
        self.class_id = class_id
        self.linked_id = linked_id

    @property
    def display_id(self):
        if self.linked_id is not None:
            return self.linked_id
        else:
            return self.track_id

    @property
    def linkable(self):
        return self.linked_id is None

    def link_to(self, track):
        self.linked_id = track.display_id

    def predict(self, motion_model=None):
        """
        Parameters
        ----------
        motion_model : if not None, predict the motion of this track given its history
        """
        if motion_model:
            self.mean, self.covariance = motion_model(self.mean, self.covariance)

        self.age += 1
        self.time_since_update += 1

        # If the track has long been missing, mark it as deleted
        if self.time_since_update > self.max_missing:
            self.state = TrackState.Deleted

    def update(self, bbx, source=None, attributes=None):
        """
        Update the state of the track. We override the predicted track position.
        Updating the track will keep or flip its state as Active
        If the confidence of detection is below the keep_alive_threshold, we will mark this track as missed.
        ----------
        bbx : new detection location of this object
        attributes: some useful attributes of this object at this frame, e.g. landmarks
        """
        self.mean = bbx[:4]
        self.confidence_score = bbx[4]
        self.time_since_update = 0
        self.state = TrackState.Active
        self.attributes = attributes
        self.source = source

        if self.confidence_score < self.keep_alive_thresh:
            self.mark_missed()

    def mark_missed(self):
        """Mark this track as missed (no association at the current time step).
        """
        # only an active track can be marked as missed
        if self.state == TrackState.Active:
            self.state = TrackState.Missing

    def is_mising(self):
        """Returns True if this track is tentative (unconfirmed).
        """
        return self.state == TrackState.Missing

    def is_active(self):
        """Returns True if this track is confirmed."""
        return self.state == TrackState.Active

    def is_deleted(self):
        """Returns True if this track is dead and should be deleted."""
        return self.state == TrackState.Deleted



"""Torch Classification Estimator"""
# pylint: disable=unused-variable,bad-whitespace,missing-function-docstring,logging-format-interpolation,arguments-differ,logging-not-lazy, not-callable
import math
import copy
import os
import logging
import time
import warnings
import pickle
from contextlib import suppress
from PIL import Image

import pandas as pd
import numpy as np
import torch
import torch.nn as nn
import torchvision.utils
from torch.optim.optimizer import Optimizer

from timm.data import create_loader, Mixup, FastCollateMixup, AugMixDataset
from timm.models import create_model, safe_model_name, convert_splitbn_model, model_parameters
from timm.utils import random_seed, dispatch_clip_grad, accuracy, unwrap_model, get_state_dict
from timm.optim import create_optimizer_v2
from timm.utils import ApexScaler, NativeScaler, ModelEmaV2, AverageMeter
from timm.loss import LabelSmoothingCrossEntropy, SoftTargetCrossEntropy

from .default import TorchImageClassificationCfg
from .utils import resolve_data_config, update_cfg, optimizer_kwargs, \
                   create_scheduler, rmse, create_optimizer_v2a
from ..constants import BINARY, MULTICLASS, REGRESSION
from ..utils import EarlyStopperOnPlateau
from ..utils import _suggest_load_context
from ..conf import _BEST_CHECKPOINT_FILE
from ..base_estimator import BaseEstimator, set_default

warnings.filterwarnings('ignore', message='.*Argument interpolation should be of type InterpolationMode instead of int.*')
logging.getLogger('timm.models').setLevel(logging.WARNING)

try:
    from apex import amp
    from apex.parallel import convert_syncbn_model
    has_apex = True
except ImportError:
    has_apex = False

has_native_amp = False
try:
    if getattr(torch.cuda.amp, 'autocast') is not None:
        has_native_amp = True
except AttributeError:
    pass


@set_default(TorchImageClassificationCfg())
class TorchImageClassificationEstimator(BaseEstimator):
    """Torch Estimator implementation for Image Classification.

    Parameters
    ----------
    config : dict
        Config in nested dict.
    logger : logging.Logger
        Optional logger for this estimator, can be `None` when default setting is used.
    reporter : callable
        The reporter for metric checkpointing.
    net : torch.nn.Module
        The custom network. If defined, the model name in config will be ignored so your
        custom network will be used for training rather than pulling it from model zoo.
    """
    def __init__(self, config, logger=None, reporter=None, net=None, optimizer=None, problem_type=None):
        super().__init__(config, logger=logger, reporter=reporter, name=None)
        if problem_type is None:
            problem_type = MULTICLASS
        self._problem_type = problem_type
        self._feature_net = None
        self._custom_net = False

        self._img_cls_cfg = self._cfg.img_cls
        self._data_cfg = self._cfg.data
        self._optimizer_cfg = self._cfg.optimizer
        self._train_cfg = self._cfg.train
        self._augmentation_cfg = self._cfg.augmentation
        self._model_ema_cfg = self._cfg.model_ema
        self._misc_cfg = self._cfg.misc

        # resolve AMP arguments based on PyTorch / Apex availability
        self.use_amp = None
        if self._misc_cfg.amp:
            # `amp` chooses native amp before apex (APEX ver not actively maintained)
            if self._misc_cfg.native_amp and has_native_amp:
                self.use_amp = 'native'
            elif self._misc_cfg.apex_amp and has_apex:
                self.use_amp = 'apex'
            elif self._misc_cfg.apex_amp or self._misc_cfg.native_amp:
                self._logger.warning(f'Neither APEX or native Torch AMP is available, using float32. \
                                       Install NVIDA apex or upgrade to PyTorch 1.6')
        # FIXME: will provided model conflict with config provided?
        if net is not None:
            assert isinstance(net, nn.Module), f"given custom network {type(net)}, `torch.nn` expected"
            try:
                net.to('cpu')
                self._custom_net = True
            except ValueError:
                pass
        self.net = net
        if optimizer is not None:
            self._logger.warning('Custom optimizer object not supported. Will follow the config instead.')
        self._optimizer = None

    def _fit(self, train_data, val_data, time_limit=math.inf):
        tic = time.time()
        self._cp_name = ''
        self._best_acc = -float('inf')
        self.epochs = self._train_cfg.epochs
        self.epoch = 0
        self.start_epoch = self._train_cfg.start_epoch
        self._time_elapsed = 0
        if max(self.start_epoch, self.epoch) >= self.epochs:
            return {'time', self._time_elapsed}
        self._init_trainer()
        self._init_model_ema()
        self._time_elapsed += time.time() - tic
        return self._resume_fit(train_data, val_data, time_limit=time_limit)

    def _resume_fit(self, train_data, val_data, time_limit=math.inf):
        tic = time.time()
        # TODO: regression not implemented
        if self._problem_type != REGRESSION and (not self.classes or not self.num_class):
            raise ValueError('This is a classification problem and we are not able to determine classes of dataset')

        if max(self.start_epoch, self.epoch) >= self.epochs:
            return {'time': self._time_elapsed}

        # wrap DP if possible
        if self.found_gpu:
            self.net = torch.nn.DataParallel(self.net, device_ids=[int(i) for i in self.valid_gpus])
        self.net = self.net.to(self.ctx[0])

        # prepare dataset
        train_dataset = train_data.to_torch()
        val_dataset = val_data.to_torch()

        # setup mixup / cutmix
        self._collate_fn = None
        self._mixup_fn = None
        self.mixup_active = self._augmentation_cfg.mixup > 0 or self._augmentation_cfg.cutmix > 0. or self._augmentation_cfg.cutmix_minmax is not None
        if self.mixup_active:
            mixup_args = dict(
                mixup_alpha=self._augmentation_cfg.mixup, cutmix_alpha=self._augmentation_cfg.cutmix,
                cutmix_minmax=self._augmentation_cfg.cutmix_minmax, prob=self._augmentation_cfg.mixup_prob,
                switch_prob=self._augmentation_cfg.mixup_switch_prob, mode=self._augmentation_cfg.mixup_mode,
                label_smoothing=self._augmentation_cfg.smoothing, num_classes=self.num_class)
            if self._misc_cfg.prefetcher:
                self._collate_fn = FastCollateMixup(**mixup_args)
            else:
                self._mixup_fn = Mixup(**mixup_args)

        # create data loaders w/ augmentation pipeiine
        train_interpolation = self._augmentation_cfg.train_interpolation
        if self._augmentation_cfg.no_aug or not train_interpolation:
            train_interpolation = self._data_cfg.interpolation
        train_loader = create_loader(
            train_dataset,
            input_size=self._data_cfg.input_size,
            batch_size=self._train_cfg.batch_size,
            is_training=True,
            use_prefetcher=self._misc_cfg.prefetcher,
            no_aug=self._augmentation_cfg.no_aug,
            scale=self._augmentation_cfg.scale,
            ratio=self._augmentation_cfg.ratio,
            hflip=self._augmentation_cfg.hflip,
            vflip=self._augmentation_cfg.vflip,
            color_jitter=self._augmentation_cfg.color_jitter,
            auto_augment=self._augmentation_cfg.auto_augment,
            interpolation=train_interpolation,
            mean=self._data_cfg.mean,
            std=self._data_cfg.std,
            num_workers=self._misc_cfg.num_workers,
            distributed=False,
            collate_fn=self._collate_fn,
            pin_memory=self._misc_cfg.pin_mem,
            use_multi_epochs_loader=self._misc_cfg.use_multi_epochs_loader
        )

        val_loader = create_loader(
            val_dataset,
            input_size=self._data_cfg.input_size,
            batch_size=self._data_cfg.validation_batch_size_multiplier * self._train_cfg.batch_size,
            is_training=False,
            use_prefetcher=self._misc_cfg.prefetcher,
            interpolation=self._data_cfg.interpolation,
            mean=self._data_cfg.mean,
            std=self._data_cfg.std,
            num_workers=self._misc_cfg.num_workers,
            distributed=False,
            crop_pct=self._data_cfg.crop_pct,
            pin_memory=self._misc_cfg.pin_mem,
        )

        self._time_elapsed += time.time() - tic
        return self._train_loop(train_loader, val_loader, time_limit=time_limit)

    def _train_loop(self, train_loader, val_loader, time_limit=math.inf):
        start_tic = time.time()
        # setup loss function
        if self.mixup_active:
            # smoothing is handled with mixup target transform
            train_loss_fn = SoftTargetCrossEntropy()
        elif self._augmentation_cfg.smoothing:
            train_loss_fn = LabelSmoothingCrossEntropy(smoothing=self._augmentation_cfg.smoothing)
        else:
            train_loss_fn = nn.CrossEntropyLoss()
        validate_loss_fn = nn.CrossEntropyLoss()
        train_loss_fn = train_loss_fn.to(self.ctx[0])
        validate_loss_fn = validate_loss_fn.to(self.ctx[0])
        eval_metric = self._misc_cfg.eval_metric
        if self._problem_type == REGRESSION:
            train_loss_fn = nn.MSELoss()
            validate_loss_fn = nn.MSELoss()
            eval_metric = 'rmse'
        early_stopper = EarlyStopperOnPlateau(
            patience=self._train_cfg.early_stop_patience,
            min_delta=self._train_cfg.early_stop_min_delta,
            baseline_value=self._train_cfg.early_stop_baseline,
            max_value=self._train_cfg.early_stop_max_value)

        self._logger.info('Start training from [Epoch %d]', max(self._train_cfg.start_epoch, self.epoch))

        self._time_elapsed += time.time() - start_tic
        for self.epoch in range(max(self.start_epoch, self.epoch), self.epochs):
            epoch = self.epoch
            if self._best_acc >= 1.0:
                self._logger.info('[Epoch {}] Early stopping as acc is reaching 1.0'.format(epoch))
                break
            should_stop, stop_message = early_stopper.get_early_stop_advice()
            if should_stop:
                self._logger.info('[Epoch {}] '.format(epoch) + stop_message)
                break
            train_metrics = self.train_one_epoch(
                epoch, self.net, train_loader, self._optimizer, train_loss_fn,
                lr_scheduler=self._lr_scheduler, output_dir=self._logdir,
                amp_autocast=self._amp_autocast, loss_scaler=self._loss_scaler, model_ema=self._model_ema, mixup_fn=self._mixup_fn, time_limit=time_limit)
            # reaching time limit, exit early
            if train_metrics['time_limit']:
                self._logger.warning(f'`time_limit={time_limit}` reached, exit early...')
                return {'train_acc': train_metrics['train_acc'], 'valid_acc': self._best_acc,
                        'time': self._time_elapsed, 'checkpoint': self._cp_name}
            post_tic = time.time()

            eval_metrics = self.validate(self.net, val_loader, validate_loss_fn, amp_autocast=self._amp_autocast)

            if self._model_ema is not None and not self._model_ema_cfg.model_ema_force_cpu:
                ema_eval_metrics = self.validate(
                    self._model_ema.module, val_loader, validate_loss_fn, amp_autocast=self._amp_autocast)
                eval_metrics = ema_eval_metrics

            if self._problem_type == REGRESSION:
                val_acc = eval_metrics['rmse']
                if self._reporter:
                    self._reporter(epoch=epoch, acc_reward=-val_acc)
                early_stopper.update(-val_acc)

                if -val_acc > self._best_acc:
                    self._cp_name = os.path.join(self._logdir, _BEST_CHECKPOINT_FILE)
                    self._logger.info('[Epoch %d] Current best rmse: %f vs previous %f, saved to %s',
                                      self.epoch, val_acc, -self._best_acc, self._cp_name)
                    self.save(self._cp_name)
                    self._best_acc = -val_acc
            else:
                val_acc = eval_metrics['top1']
                if self._reporter:
                    self._reporter(epoch=epoch, acc_reward=val_acc)
                early_stopper.update(val_acc)

                if val_acc > self._best_acc:
                    self._cp_name = os.path.join(self._logdir, _BEST_CHECKPOINT_FILE)
                    self._logger.info('[Epoch %d] Current best top-1: %f vs previous %f, saved to %s',
                                      self.epoch, val_acc, self._best_acc, self._cp_name)
                    self.save(self._cp_name)
                    self._best_acc = val_acc

            if self._lr_scheduler is not None:
                # step LR for next epoch
                self._lr_scheduler.step(epoch + 1, eval_metrics[eval_metric])

            self._time_elapsed += time.time() - post_tic

        if 'accuracy' in train_metrics:
            return {'train_acc': train_metrics['accuracy'], 'valid_acc': self._best_acc,
                    'time': self._time_elapsed, 'checkpoint': self._cp_name}
        # rmse
        else:
            if self._problem_type == REGRESSION:
                return {'train_score': train_metrics['rmse'], 'valid_score': -self._best_acc,
                        'time': self._time_elapsed, 'checkpoint': self._cp_name}
            # mixup
            else:
                return {'train_score': train_metrics['rmse'], 'valid_acc': self._best_acc,
                        'time': self._time_elapsed, 'checkpoint': self._cp_name}

    def train_one_epoch(
            self, epoch, net, loader, optimizer, loss_fn,
            lr_scheduler=None, output_dir=None, amp_autocast=suppress,
            loss_scaler=None, model_ema=None, mixup_fn=None, time_limit=math.inf):
        start_tic = time.time()
        if self._augmentation_cfg.mixup_off_epoch and epoch >= self._augmentation_cfg.mixup_off_epoch:
            if self._misc_cfg.prefetcher and loader.mixup_enabled:
                loader.mixup_enabled = False
            elif mixup_fn is not None:
                mixup_fn.mixup_enabled = False

        second_order = hasattr(optimizer, 'is_second_order') and optimizer.is_second_order
        losses_m = AverageMeter()
        train_metric_score_m = AverageMeter()

        net.train()

        num_updates = epoch * len(loader)
        self._time_elapsed += time.time() - start_tic
        tic = time.time()
        last_tic = time.time()
        train_metric_name = 'accuracy'
        batch_idx = 0
        for batch_idx, (input, target) in enumerate(loader):
            b_tic = time.time()
            if self._time_elapsed > time_limit:
                return {'train_acc': train_metric_score_m.avg, 'train_loss': losses_m.avg, 'time_limit': True}
            if self._problem_type == REGRESSION:
                target = target.to(torch.float32)
            if not self._misc_cfg.prefetcher:
                # prefetcher would move data to cuda by default
                input, target = input.to(self.ctx[0]), target.to(self.ctx[0])
                if mixup_fn is not None:
                    input, target = mixup_fn(input, target)

            with amp_autocast():
                output = net(input)
                if self._problem_type == REGRESSION:
                    output = output.flatten()
                loss = loss_fn(output, target)
            if self._problem_type == REGRESSION:
                train_metric_name = 'rmse'
                train_metric_score = rmse(output, target)
            else:
                if output.shape == target.shape:
                    train_metric_name = 'rmse'
                    train_metric_score = rmse(output, target)
                else:
                    train_metric_score = accuracy(output, target)[0] / 100

            losses_m.update(loss.item(), input.size(0))
            train_metric_score_m.update(train_metric_score.item(), output.size(0))

            optimizer.zero_grad()
            if loss_scaler is not None:
                loss_scaler(
                    loss, optimizer,
                    clip_grad=self._optimizer_cfg.clip_grad, clip_mode=self._optimizer_cfg.clip_mode,
                    parameters=model_parameters(net, exclude_head='agc' in self._optimizer_cfg.clip_mode),
                    create_graph=second_order)
            else:
                loss.backward(create_graph=second_order)
                if self._optimizer_cfg.clip_grad is not None:
                    dispatch_clip_grad(
                        model_parameters(net, exclude_head='agc' in self._optimizer_cfg.clip_mode),
                        value=self._optimizer_cfg.clip_grad, mode=self._optimizer_cfg.clip_mode)
                optimizer.step()

            if model_ema is not None:
                model_ema.update(net)

            if self.found_gpu:
                torch.cuda.synchronize()

            num_updates += 1
            if (batch_idx+1) % self._misc_cfg.log_interval == 0:
                lrl = [param_group['lr'] for param_group in optimizer.param_groups]
                lr = sum(lrl) / len(lrl)
                self._logger.info('Epoch[%d] Batch [%d]\tSpeed: %f samples/sec\t%s=%f\tlr=%f',
                                  epoch, batch_idx,
                                  self._train_cfg.batch_size*self._misc_cfg.log_interval/(time.time()-last_tic),
                                  train_metric_name, train_metric_score_m.avg, lr)
                last_tic = time.time()

                if self._misc_cfg.save_images and output_dir:
                    torchvision.utils.save_image(
                        input,
                        os.path.join(output_dir, 'train-batch-%d.jpg' % batch_idx),
                        padding=0,
                        normalize=True)

            if lr_scheduler is not None:
                lr_scheduler.step_update(num_updates=num_updates, metric=losses_m.avg)

            self._time_elapsed += time.time() - b_tic

        throughput = int(self._train_cfg.batch_size * batch_idx / (time.time() - tic))
        self._logger.info('[Epoch %d] training: %s=%f', epoch, train_metric_name, train_metric_score_m.avg)
        self._logger.info('[Epoch %d] speed: %d samples/sec\ttime cost: %f', epoch, throughput, time.time()-tic)

        end_time = time.time()
        if hasattr(optimizer, 'sync_lookahead'):
            optimizer.sync_lookahead()

        self._time_elapsed += time.time() - end_time

        return {train_metric_name: train_metric_score_m.avg, 'train_loss': losses_m.avg, 'time_limit': False}

    def validate(self, net, loader, loss_fn, amp_autocast=suppress, metric_name=None):
        losses_m = AverageMeter()
        top1_m = AverageMeter()
        top5_m = AverageMeter()
        rmse_m = AverageMeter()

        net.eval()

        with torch.no_grad():
            for batch_idx, (input, target) in enumerate(loader):
                if not self._misc_cfg.prefetcher:
                    input = input.to(self.ctx[0])
                    target = target.to(self.ctx[0])

                with amp_autocast():
                    output = net(input)
                    if self._problem_type == REGRESSION:
                        output = output.flatten()
                if isinstance(output, (tuple, list)):
                    output = output[0]

                if self._problem_type == REGRESSION:
                    if metric_name:
                        assert metric_name == 'rmse', f'{metric_name} metric not supported for regression.'
                    val_metric_score = rmse(output, target)
                else:
                    val_metric_score = accuracy(output, target, topk=(1, min(5, self.num_class)))

                # augmentation reduction
                reduce_factor = self._misc_cfg.tta
                if self._problem_type != REGRESSION and reduce_factor > 1:
                    output = output.unfold(0, reduce_factor, reduce_factor).mean(dim=2)
                    target = target[0:target.size(0):reduce_factor]

                loss = loss_fn(output, target)
                reduced_loss = loss.data

                if self.found_gpu:
                    torch.cuda.synchronize()

                losses_m.update(reduced_loss.item(), input.size(0))
                if self._problem_type == REGRESSION:
                    rmse_score = val_metric_score
                    rmse_m.update(rmse_score.item(), output.size(0))
                else:
                    acc1, acc5 = val_metric_score
                    acc1 /= 100
                    acc5 /= 100
                    top1_m.update(acc1.item(), output.size(0))
                    top5_m.update(acc5.item(), output.size(0))

        if self._problem_type == REGRESSION:
            self._logger.info('[Epoch %d] validation: rmse=%f', self.epoch, rmse_m.avg)
            return {'loss': losses_m.avg, 'rmse': rmse_m.avg}
        else:
            self._logger.info('[Epoch %d] validation: top1=%f top5=%f', self.epoch, top1_m.avg, top5_m.avg)
            return {'loss': losses_m.avg, 'top1': top1_m.avg, 'top5': top5_m.avg}

    def _init_network(self, **kwargs):
        load_only = kwargs.get('load_only', False)
        if not self.num_class and self._problem_type != REGRESSION:
            raise ValueError('This is a classification problem and we are not able to create network when `num_class` is unknown. \
                It should be inferred from dataset or resumed from saved states.')
        assert len(self.classes) == self.num_class

        # Disable syncBatchNorm as it's only supported on DDP
        if self._train_cfg.sync_bn:
            self._logger.info(
                'Disable Sync batch norm as it is not supported for now.')
            update_cfg(self._cfg, {'train': {'sync_bn': False}})

        # ctx
        self.found_gpu = False
        valid_gpus = []
        if self._cfg.gpus:
            valid_gpus = self._torch_validate_gpus(self._cfg.gpus)
            self.found_gpu = True
            if not valid_gpus:
                self.found_gpu = False
                self._logger.warning(
                    'No gpu detected, fallback to cpu. You can ignore this warning if this is intended.')
            elif len(valid_gpus) != len(self._cfg.gpus):
                self._logger.warning(
                    f'Loaded on gpu({valid_gpus}), different from gpu({self._cfg.gpus}).')
        self.ctx = [torch.device(f'cuda:{gid}') for gid in valid_gpus] if self.found_gpu else [torch.device('cpu')]
        self.valid_gpus = valid_gpus

        if not self.found_gpu and self.use_amp:
            self.use_amp = None
            self._logger.warning('Training on cpu. AMP disabled.')
            update_cfg(self._cfg, {'misc': {'amp': False, 'apex_amp': False, 'native_amp': False}})

        if not self.found_gpu and self._misc_cfg.prefetcher:
            self._logger.warning(
                'Training on cpu. Prefetcher disabled.')
            update_cfg(self._cfg, {'misc': {'prefetcher': False}})
            self._logger.warning(
                'Training on cpu. SyncBatchNorm disabled.')
            update_cfg(self._cfg, {'train': {'sync_bn': False}})

        random_seed(self._misc_cfg.seed)

        if not self.net:
            self.net = create_model(
                self._img_cls_cfg.model,
                pretrained=self._img_cls_cfg.pretrained and not load_only,
                num_classes=max(self.num_class, 1),
                global_pool=self._img_cls_cfg.global_pool_type,
                drop_rate=self._augmentation_cfg.drop,
                drop_path_rate=self._augmentation_cfg.drop_path,
                drop_block_rate=self._augmentation_cfg.drop_block,
                bn_momentum=self._train_cfg.bn_momentum,
                bn_eps=self._train_cfg.bn_eps,
                scriptable=self._misc_cfg.torchscript
            )

            self._logger.info(f'Model {safe_model_name(self._img_cls_cfg.model)} created, param count: \
                                        {sum([m.numel() for m in self.net.parameters()])}')
        else:
            self._logger.info(f'Use user provided model. Neglect model in config.')
            out_features = list(self.net.children())[-1].out_features
            if self._problem_type != REGRESSION:
                assert out_features == self.num_class, f'Custom model out_feature {out_features} != num_class {self.num_class}.'
            else:
                assert out_features == 1, f'Regression problem expects num_out_feature == 1, got {out_features} instead.'

        resolve_data_config(self._cfg, model=self.net)

        self.net = self.net.to(self.ctx[0])

        # setup synchronized BatchNorm
        if self._train_cfg.sync_bn:
            if has_apex and self.use_amp != 'native':
                # Apex SyncBN preferred unless native amp is activated
                self.net = convert_syncbn_model(self.net)
            else:
                self.net = torch.nn.SyncBatchNorm.convert_sync_batchnorm(self.net)
            self._logger.info(
                'Converted model to use Synchronized BatchNorm. WARNING: You may have issues if using '
                'zero initialized BN layers (enabled by default for ResNets) while sync-bn enabled.')

        if self._misc_cfg.torchscript:
            assert not self.use_amp == 'apex', 'Cannot use APEX AMP with torchscripted model'
            assert not self._train_cfg.sync_bn, 'Cannot use SyncBatchNorm with torchscripted model'
            self.net = torch.jit.script(self.net)

    def _init_trainer(self):
        if self._optimizer is None:
            if self._img_cls_cfg.pretrained and not self._custom_net \
                and (self._train_cfg.transfer_lr_mult != 1 or self._train_cfg.output_lr_mult != 1):
                # adjust feature/last_fc learning rate multiplier in optimizer
                self._logger.debug(f'Reduce network lr multiplier to {self._train_cfg.transfer_lr_mult}, while keep ' +
                                   f'last FC layer lr_mult to {self._train_cfg.output_lr_mult}')
                optim_kwargs = optimizer_kwargs(cfg=self._cfg)
                optim_kwargs['feature_lr_mult'] = self._cfg.train.transfer_lr_mult
                optim_kwargs['fc_lr_mult'] = self._cfg.train.output_lr_mult
                self._optimizer = create_optimizer_v2(self.net, **optimizer_kwargs(cfg=self._cfg))
            else:
                self._optimizer = create_optimizer_v2(self.net, **optimizer_kwargs(cfg=self._cfg))
        self._init_loss_scaler()
        self._lr_scheduler, self.epochs = create_scheduler(self._cfg, self._optimizer)
        self._lr_scheduler.step(self.start_epoch, self.epoch)

    def _init_loss_scaler(self):
        # setup automatic mixed-precision (AMP) loss scaling and op casting
        self._amp_autocast = suppress  # do nothing
        self._loss_scaler = None
        if self.use_amp == 'apex':
            self.net, self._optimizer = amp.initialize(self.net, self._optimizer, opt_level='O1')
            self._loss_scaler = ApexScaler()
            self._logger.info('Using NVIDIA APEX AMP. Training in mixed precision.')
        elif self.use_amp == 'native':
            self._amp_autocast = torch.cuda.amp.autocast
            self._loss_scaler = NativeScaler()
            self._logger.info('Using native Torch AMP. Training in mixed precision.')
        else:
            self._logger.info('AMP not enabled. Training in float32.')

    def _init_model_ema(self):
        # Disable for now
        if self._model_ema_cfg.model_ema:
            self._logger.info('Disable EMA as it is not supported for now.')
            update_cfg(self._cfg, {'model_ema': {'model_ema': False}})
        # setup exponential moving average of model weights, SWA could be used here too
        self._model_ema = None
        if self._model_ema_cfg.model_ema:
            # Important to create EMA model after cuda(), DP wrapper, and AMP but before SyncBN and DDP wrapper
            self._model_ema = ModelEmaV2(
                self.net, decay=self._model_ema_cfg.model_ema_decay, device='cpu' if self._model_ema_cfg.model_ema_force_cpu else None)


    def evaluate(self, val_data, metric_name=None):
        return self._evaluate(val_data, metric_name)

    def _evaluate(self, val_data, metric_name=None):
        if self._problem_type == REGRESSION:
            validate_loss_fn = nn.MSELoss()
        else:
            validate_loss_fn = nn.CrossEntropyLoss()
        validate_loss_fn = validate_loss_fn.to(self.ctx[0])
        val_data = val_data.to_torch()
        val_loader = create_loader(
            val_data,
            input_size=self._data_cfg.input_size,
            batch_size=self._data_cfg.validation_batch_size_multiplier * self._train_cfg.batch_size,
            is_training=False,
            use_prefetcher=self._misc_cfg.prefetcher,
            interpolation=self._data_cfg.interpolation,
            mean=self._data_cfg.mean,
            std=self._data_cfg.std,
            num_workers=self._misc_cfg.num_workers,
            distributed=False,
            crop_pct=self._data_cfg.crop_pct,
            pin_memory=self._misc_cfg.pin_mem,
        )
        return self.validate(self.net, val_loader, validate_loss_fn, amp_autocast=self._amp_autocast, metric_name=metric_name)

    def _predict(self, x, **kwargs):
        with_proba = kwargs.get('with_proba', False)
        if with_proba and self._problem_type not in [MULTICLASS, BINARY]:
            raise AssertionError('with_proba is only supported for classification problems. Please use predict instead.')
        if isinstance(x, str):
            return self._predict((x,), **kwargs).drop(columns=['image'], errors='ignore')
        elif isinstance(x, pd.DataFrame):
            assert 'image' in x.columns, "Expect column `image` for input images"
            df = self._predict(tuple(x['image']), **kwargs)
            return df.reset_index(drop=True)
        elif isinstance(x, (list, tuple)):
            loader = create_loader(
                ImageListDataset(x),
                input_size=self._data_cfg.input_size,
                batch_size=self._train_cfg.batch_size,
                use_prefetcher=self._misc_cfg.prefetcher,
                interpolation=self._data_cfg.interpolation,
                mean=self._data_cfg.mean,
                std=self._data_cfg.std,
                num_workers=self._misc_cfg.num_workers,
                crop_pct=self._data_cfg.crop_pct
            )

            self.net.eval()

            topk = min(5, self.num_class)
            results = []
            idx = 0
            with torch.no_grad():
                for input, _ in loader:
                    input = input.to(self.ctx[0])
                    labels = self.net(input)
                    for l in labels:
                        if self._problem_type in [MULTICLASS, BINARY]:
                            probs = nn.functional.softmax(l, dim=0).cpu().numpy().flatten()
                            if with_proba:
                                results.append({'image_proba': probs.tolist(), 'image': x[idx]})
                            else:
                                topk_inds = l.topk(topk)[1].cpu().numpy().flatten()
                                results.extend([{'class': self.classes[topk_inds[k]],
                                                 'score': probs[topk_inds[k]],
                                                 'id': topk_inds[k],
                                                 'image': x[idx]}
                                                for k in range(topk)])
                        else:
                            results.append({'prediction': l.cpu().numpy().flatten(), 'image': x[idx]})
                        idx += 1
            return pd.DataFrame(results)
        elif not isinstance(x, torch.Tensor):
            raise ValueError('Input is not supported: {}'.format(type(x)))
        assert len(x.shape) == 4 and x.shape[1] == 3, f"Expect input to be (n, 3, h, w), given {x.shape}"
        with torch.no_grad():
            input = x.to(self.ctx[0])
            label = self.net(input)
            if self._problem_type in [MULTICLASS, BINARY]:
                topk = min(5, self.num_class)
                probs = nn.functional.softmax(label, dim=1).cpu().numpy().flatten()
                topk_inds = label.topk(topk)[1].cpu().numpy().flatten()
                if with_proba:
                    df = pd.DataFrame([{'image_proba': probs.tolist()}])
                else:
                    df = pd.DataFrame([{'class': self.classes[topk_inds[k]],
                                        'score': probs[topk_inds[k]],
                                        'id': topk_inds[k]}
                                       for k in range(topk)])
            else:
                df = pd.DataFrame([{'prediction': label.cpu().numpy().flatten()}])
        return df


    def _predict_feature(self, x, **kwargs):
        if isinstance(x, str):
            return self._predict_feature((x,))
        elif isinstance(x, pd.DataFrame):
            assert 'image' in x.columns, "Expect column `image` for input images"
            df = self._predict_feature(tuple(x['image']))
            df = df.set_index(x.index)
            df['image'] = x['image']
            return df
        elif isinstance(x, (list, tuple)):
            assert isinstance(x[0], str), "expect image paths in list/tuple input"
            loader = create_loader(
                ImageListDataset(x),
                input_size=self._data_cfg.input_size,
                batch_size=self._train_cfg.batch_size,
                use_prefetcher=self._misc_cfg.prefetcher,
                interpolation=self._data_cfg.interpolation,
                mean=self._data_cfg.mean,
                std=self._data_cfg.std,
                num_workers=self._misc_cfg.num_workers,
                crop_pct=self._data_cfg.crop_pct
            )

            self.net.eval()

            results = []
            with torch.no_grad():
                for input, _ in loader:
                    input = input.to(self.ctx[0])
                    try:
                        features = self.net.forward_features(input)
                    except AttributeError:
                        features = self.net.module.forward_features(input)
                    for f in features:
                        f = f.cpu().numpy().flatten()
                        results.append({'image_feature': f})
            df = pd.DataFrame(results)
            df['image'] = x
            return df
        elif not isinstance(x, torch.Tensor):
            raise ValueError('Input is not supported: {}'.format(type(x)))
        with torch.no_grad():
            input = x.to(self.ctx[0])
            feature = self.net.forward_features(input)
            result = [{'image_feature': feature}]
        df = pd.DataFrame(result)
        return df

    def _reconstruct_state_dict(self, state_dict):
        new_state_dict = {}
        for k, v in state_dict.items():
            name = k[7:] if k.startswith('module') else k
            new_state_dict[name] = v
        return new_state_dict

    def save(self, filename):
        d = dict()
        current_states = self.__dict__.copy()
        if self.net:
            if not self._custom_net:
                if isinstance(self.net, torch.nn.DataParallel):
                    d['model_state_dict'] = get_state_dict(self.net.module, unwrap_model)
                else:
                    d['model_state_dict'] = get_state_dict(self.net, unwrap_model)
            else:
                net_pickle = pickle.dumps(self.net)
                d['net_pickle'] = net_pickle
            self.net = None
        if self._optimizer:
            d['optimizer_state_dict'] = self._optimizer.state_dict()
            self._optimizer = None
        if hasattr(self, '_loss_scaler') and self._loss_scaler:
            d[self._loss_scaler.state_dict_key] = self._loss_scaler.state_dict()
            d['_loss_scaler_state_dict_key'] = self._loss_scaler.state_dict_key
        if self._model_ema:
            d['ema_state_dict'] = get_state_dict(self._model_ema, unwrap_model)
            self._model_ema = None
        self._logger = None
        self._reporter = None
        d['estimator'] = self
        torch.save(d, filename)
        self.__dict__.update(current_states)

    @classmethod
    def load(cls, filename, ctx='auto'):
        d = torch.load(filename, map_location=torch.device('cpu'))
        est = d.pop('estimator')
        # logger
        est._logger = logging.getLogger(cls.__name__)
        est._logger.setLevel(logging.ERROR)
        try:
            fh = logging.FileHandler(est._log_file)
            est._logger.addHandler(fh)
        #pylint: disable=bare-except
        except:
            pass
        model_state_dict = d.get('model_state_dict', None)
        net_pickle = d.get('net_pickle', None)
        if model_state_dict:
            est._init_network(load_only=True)
            net_state_dict = est._reconstruct_state_dict(model_state_dict)
            if isinstance(est.net, torch.nn.DataParallel):
                est.net.module.load_state_dict(net_state_dict)
            else:
                est.net.load_state_dict(net_state_dict)
        elif net_pickle:
            est.net = pickle.loads(net_pickle)
        optimizer_state_dict = d.get('optimizer_state_dict', None)
        if optimizer_state_dict:
            est._init_trainer()
            est._optimizer.load_state_dict(optimizer_state_dict)
        if hasattr(est, '_loss_scaler') and est._loss_scaler:
            loss_scaler_state_dict_key = d.get('loss_scaler_state_dict')
            loss_scaler_dict = d.get(loss_scaler_state_dict_key, None)
            if loss_scaler_dict:
                est._loss_scaler.load_state_dict(loss_scaler_dict)
        ema_state_dict = d.get('ema_state_dict', None)
        est._init_model_ema()
        if ema_state_dict:
            ema_state_dict = est._reconstruct_state_dict(ema_state_dict)
            if isinstance(est.net, torch.nn.DataParallel):
                est._model_ema.module.module.load_state_dict(ema_state_dict)
            else:
                est._model_ema.module.load_state_dict(ema_state_dict)
        new_ctx = _suggest_load_context(est.net, ctx, est.ctx)
        est.reset_ctx(new_ctx)
        est._logger.setLevel(logging.INFO)
        return est

    # pylint: disable=redefined-outer-name, reimported
    def __getstate__(self):
        d = self.__dict__.copy()
        try:
            import torch
            net = d.pop('net', None)
            model_ema = d.pop('_model_ema', None)
            optimizer = d.pop('_optimizer', None)
            loss_scaler = d.pop('_loss_scaler', None)
            save_state = {}
            if net is not None:
                if not self._custom_net:
                    if isinstance(net, torch.nn.DataParallel):
                        save_state['state_dict'] = get_state_dict(net.module, unwrap_model)
                    else:
                        save_state['state_dict'] = get_state_dict(net, unwrap_model)
                else:
                    net_pickle = pickle.dumps(net)
                    save_state['net_pickle'] = net_pickle
            if optimizer is not None:
                save_state['optimizer'] = optimizer.state_dict()
            if loss_scaler is not None:
                save_state[loss_scaler.state_dict_key] = loss_scaler.state_dict()
            if model_ema is not None:
                save_state['state_dict_ema'] = get_state_dict(model_ema, unwrap_model)
        except ImportError:
            pass
        d['save_state'] = save_state
        d['_logger'] = None
        d['_reporter'] = None
        return d

    def __setstate__(self, state):
        save_state = state.pop('save_state', None)
        self.__dict__.update(state)
        # logger
        self._logger = logging.getLogger(state.get('_name', self.__class__.__name__))
        self._logger.setLevel(logging.ERROR)
        try:
            fh = logging.FileHandler(self._log_file)
            self._logger.addHandler(fh)
        #pylint: disable=bare-except
        except:
            pass
        if not save_state:
            self.net = None
            self._optimizer = None
            self._logger.setLevel(logging.INFO)
            return
        try:
            import torch
            self.net = None
            self._optimizer = None
            if self._custom_net:
                if save_state.get('net_pickle', None):
                    self.net = pickle.loads(save_state['net_pickle'])
            else:
                if save_state.get('state_dict', None):
                    self._init_network(load_only=True)
                    net_state_dict = self._reconstruct_state_dict(save_state['state_dict'])
                    if isinstance(self.net, torch.nn.DataParallel):
                        self.net.module.load_state_dict(net_state_dict)
                    else:
                        self.net.load_state_dict(net_state_dict)
            if save_state.get('optimizer', None):
                self._init_trainer()
                self._optimizer.load_state_dict(save_state['optimizer'])
            if hasattr(self, '_loss_scaler') and self._loss_scaler and self._loss_scaler.state_dict_key in save_state:
                loss_scaler_dict = save_state[self._loss_scaler.state_dict_key]
                self._loss_scaler.load_state_dict(loss_scaler_dict)
            if save_state.get('state_dict_ema', None):
                self._init_model_ema()
                model_ema_dict = save_state.get('state_dict_ema')
                model_ema_dict = self._reconstruct_state_dict(model_ema_dict)
                if isinstance(self.net, torch.nn.DataParallel):
                    self._model_ema.module.module.load_state_dict(model_ema_dict)
                else:
                    self._model_ema.module.load_state_dict(model_ema_dict)
        except ImportError:
            pass
        self._logger.setLevel(logging.INFO)


class ImageListDataset(torch.utils.data.Dataset):
    """An internal image list dataset for batch predict"""
    def __init__(self, imlist):
        self._imlist = imlist
        self.transform = None

    def __getitem__(self, idx):
        img = Image.open(self._imlist[idx]).convert('RGB')
        label = None
        if self.transform is not None:
            img = self.transform(img)
        return img, torch.tensor(-1, dtype=torch.long)

    def __len__(self):
        return len(self._imlist)



"""Multigrid support to speed up training of video models"""
from functools import reduce
import numpy as np

from torch.utils.data import Sampler
from torch._six import int_classes as _int_classes


__all__ = ['multiGridHelper', 'MultiGridBatchSampler']


sq2 = np.sqrt(2)
class multiGridHelper(object):
    """
    A Multigrid Method for Efficiently Training Video Models
    Chao-Yuan Wu, Ross Girshick, Kaiming He, Christoph Feichtenhofer, Philipp Krähenbühl
    CVPR 2020, https://arxiv.org/abs/1912.00998
    """
    def __init__(self):
        # Scale: [T, H, W]
        self.long_cycle = np.asarray([[1, 1, 1], [2, 1, 1], [2, sq2, sq2]])[::-1]
        self.short_cycle = np.asarray([[1, 1, 1], [1, sq2, sq2], [1, 2, 2]])[::-1]
        self.short_cycle_sp = np.asarray([[1, 1, 1], [1, sq2, sq2], [1, sq2, sq2]])[::-1]
        self.mod_long = len(self.long_cycle)
        self.mod_short = len(self.short_cycle)

    def scale(self, x):
        return int(np.around(reduce(lambda a, b: a * b, x)))

    def get_scale(self, alpha, beta):
        long_scale = self.scale(self.long_cycle[alpha])
        if alpha == 0:
            short_scale = self.scale(self.short_cycle_sp[beta])
        else:
            short_scale = self.scale(self.short_cycle[beta])
        return long_scale * short_scale

    def get_scale_alpha(self, alpha):
        long_scale = self.scale(self.long_cycle[alpha])
        return long_scale

    def get_scale_beta(self, beta):
        short_scale = self.scale(self.short_cycle[beta])
        return short_scale

    def get_resize(self, alpha, beta):
        long_item = self.long_cycle[alpha]
        short_item = self.short_cycle[beta]
        return long_item * short_item


class MultiGridBatchSampler(Sampler):
    """
    A Multigrid Method for Efficiently Training Video Models
    Chao-Yuan Wu, Ross Girshick, Kaiming He, Christoph Feichtenhofer, Philipp Krähenbühl
    CVPR 2020, https://arxiv.org/abs/1912.00998
    """
    def __init__(self, sampler, batch_size, drop_last, use_long=False, use_short=False):
        '''
        :param sampler: torch.utils.data.Sample
        :param batch_size: int
        :param drop_last: bool
        :param use_long: bool
        :param use_short: bool
        Apply batch collecting function based on multiGridHelper definition
        '''
        if not isinstance(sampler, Sampler):
            raise ValueError("sampler should be an instance of "
                             "torch.utils.data.Sampler, but got sampler={}"
                             .format(sampler))
        if not isinstance(batch_size, _int_classes) or isinstance(batch_size, bool) or \
                batch_size <= 0:
            raise ValueError("batch_size should be a positive integer value, "
                             "but got batch_size={}".format(batch_size))
        if not isinstance(drop_last, bool):
            raise ValueError("drop_last should be a boolean value, but got "
                             "drop_last={}".format(drop_last))
        if not isinstance(use_long, bool):
            raise ValueError("use_long should be a boolean value, but got "
                             "use_long={}".format(use_long))
        if not isinstance(use_short, bool):
            raise ValueError("use_short should be a boolean value, but got "
                             "use_short={}".format(use_short))
        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last

        self.mg_helper = multiGridHelper()
        # single grid setting
        self.alpha = self.mg_helper.mod_long - 1
        self.beta = self.mg_helper.mod_short - 1
        self.short_cycle_label = False
        if use_long:
            self.activate_long_cycle()
        if use_short:
            self.activate_short_cycle()
        self.batch_scale = self.mg_helper.get_scale(self.alpha, self.beta)

    def activate_short_cycle(self):
        self.short_cycle_label = True
        self.beta = 0

    def activate_long_cycle(self):
        self.alpha = 0

    def deactivate(self):
        self.alpha = self.mg_helper.mod_long - 1
        self.beta = self.mg_helper.mod_short - 1
        self.short_cycle_label = False

    def __iter__(self):
        batch = []
        if self.short_cycle_label:
            self.beta = 0
        else:
            self.beta = self.mg_helper.mod_short - 1
        self.batch_scale = self.mg_helper.get_scale(self.alpha, self.beta)
        for idx in self.sampler:
            batch.append([idx, self.alpha, self.beta])
            if len(batch) == self.batch_size*self.batch_scale:
                yield batch
                batch = []
                if self.short_cycle_label:
                    self.beta = (self.beta + 1) % self.mg_helper.mod_short
                    self.batch_scale = self.mg_helper.get_scale(self.alpha, self.beta)

        if len(batch) > 0 and not self.drop_last:
            yield batch

    def step_long_cycle(self):
        self.alpha = (self.alpha + 1) % self.mg_helper.mod_long

    # def compute_lr_milestone(self, lr_milestone):
    #     """
    #     long cycle milestones, deprecated. Define long cycle in config files
    #     """
    #     self.len_long = self.mg_helper.mod_long
    #     self.n_epoch_long = 0
    #     for x in range(self.len_long):
    #         self.n_epoch_long += self.mg_helper.get_scale_alpha(x)
    #     lr_long_cycle = []
    #     for i, _ in enumerate(lr_milestone):
    #         if i == 0:
    #             pre = 0
    #         else:
    #             pre = lr_milestone[i-1]
    #         cycle_length = (lr_milestone[i] - pre) // self.n_epoch_long
    #         bonus = (lr_milestone[i] - pre)%self.n_epoch_long // self.len_long
    #         for j in range(self.len_long)[::-1]:
    #             pre = pre + cycle_length*(2**j) + bonus
    #             if j == 0:
    #                 pre = lr_milestone[i]
    #             lr_long_cycle.append(pre)
    #     lr_long_cycle.append(0)
    #     lr_long_cycle = sorted(lr_long_cycle)
    #     return lr_long_cycle

    def __len__(self):
        scale_per_short_cycle = 0
        for x in range(self.mg_helper.mod_short):
            scale_per_short_cycle += self.mg_helper.get_scale(self.alpha, x)
        num_full_short_cycle = len(self.sampler) // (self.batch_size * scale_per_short_cycle)

        total = num_full_short_cycle * self.mg_helper.mod_short
        remain = len(self.sampler) % (self.batch_size * scale_per_short_cycle)
        for x in range(self.mg_helper.mod_short):
            remain = remain - self.mg_helper.get_scale(self.alpha, x)*self.batch_size
            if remain >= 0 or (remain < 0 and self.drop_last is False):
                total += 1
            if remain <= 0:
                break
        assert remain <= 0
        return total



"""Estimator implementations"""
from .utils import create_dummy_estimator
# FIXME: for quick test purpose only
try:
    import mxnet
    from .image_classification import ImageClassificationEstimator
    from .ssd import SSDEstimator
    from .yolo import YOLOv3Estimator
    from .faster_rcnn import FasterRCNNEstimator
    # from .mask_rcnn import MaskRCNNEstimator
    from .center_net import CenterNetEstimator
except ImportError:
    # create dummy placeholder estimator classes
    reason = 'gluoncv.auto.estimators.{} requires mxnet to be installed which is missing.'
    ImageClassificationEstimator = create_dummy_estimator(
        'ImageClassificationEstimator', reason)
    SSDEstimator = create_dummy_estimator(
        'SSDEstimator', reason)
    YOLOv3Estimator = create_dummy_estimator(
        'YOLOv3Estimator', reason)
    FasterRCNNEstimator = create_dummy_estimator(
        'FasterRCNNEstimator', reason)
    CenterNetEstimator = create_dummy_estimator(
        'CenterNetEstimator', reason)

try:
    import timm
    import torch
    from .torch_image_classification import TorchImageClassificationEstimator
except ImportError:
    reason = 'This estimator requires torch/timm to be installed which is missing.'
    TorchImageClassificationEstimator = create_dummy_estimator(
        'TorchImageClassificationEstimator', reason)



# pylint: disable=missing-function-docstring, line-too-long, consider-using-with
"""
Utility functions, misc
"""
import os
import sys
import time
import functools
import logging
from contextlib import contextmanager
from functools import wraps

import numpy as np
import torch.nn as nn
import torch


def read_labelmap(labelmap_file):
    """Read label map and class ids."""

    labelmap = []
    class_ids = set()
    name = ""
    class_id = ""
    with open(labelmap_file, 'r') as f:
        for line in f:
            if line.startswith("  name:"):
                name = line.split('"')[1]
            elif line.startswith("  id:") or line.startswith("  label_id:"):
                class_id = int(line.strip().split(" ")[-1])
                labelmap.append({"id": class_id, "name": name})
                class_ids.add(class_id)
    return labelmap, class_ids

# cache the opened file object, so that different calls to `setup_logger`
# with the same file name can safely write to the same file.
@functools.lru_cache(maxsize=None)
def _cached_log_stream(filename):
    return open(filename, "a")

def build_log_dir(cfg):
    # create base log directory
    if cfg.CONFIG.LOG.EXP_NAME == 'use_time':
        cfg.CONFIG.LOG.EXP_NAME = time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime(time.time()))
    log_path = os.path.join(cfg.CONFIG.LOG.BASE_PATH, cfg.CONFIG.LOG.EXP_NAME)
    if not os.path.exists(log_path):
        os.makedirs(log_path)

    # setup log file and stdout
    logger = logging.getLogger("gluoncv")
    log_level = logging.getLevelName(cfg.CONFIG.LOG.LEVEL)
    plain_formatter = logging.Formatter(
        "[%(asctime)s] %(name)s %(levelname)s: %(message)s", datefmt="%m/%d %H:%M:%S"
    )
    logger.setLevel(log_level)
    logger.propagate = False
    # file logger, enabled for all workers
    log_filename = os.path.join(log_path, cfg.CONFIG.LOG.LOG_FILENAME)
    if cfg.DDP_CONFIG.WORLD_RANK > 0:
        log_filename += f".rank{cfg.DDP_CONFIG.WORLD_RANK}"
    fh = logging.StreamHandler(_cached_log_stream(log_filename))
    fh.setLevel(log_level)
    fh.setFormatter(plain_formatter)
    logger.addHandler(fh)
    if cfg.DDP_CONFIG.WORLD_RANK == 0:
        # stdout in rank 0 only
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(log_level)
        logger.addHandler(ch)

    # dump config file
    with open(os.path.join(log_path, 'config.yaml'), 'w') as f:
        f.write(cfg.dump())

    # create tensorboard saving directory
    tb_logdir = os.path.join(log_path, cfg.CONFIG.LOG.LOG_DIR)
    if not os.path.exists(tb_logdir):
        os.makedirs(tb_logdir)

    # create checkpoint saving directory
    ckpt_logdir = os.path.join(log_path, cfg.CONFIG.LOG.SAVE_DIR)
    if not os.path.exists(ckpt_logdir):
        os.makedirs(ckpt_logdir)

    return tb_logdir


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def calculate_mAP(output, target):
    import torchnet.meter as meter
    mtr = meter.mAPMeter()
    mtr.add(output, target)
    ap = mtr.value()
    return ap


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


def per_class_error(output, target, num_classes):
    per_class_acc = []
    for i in range(num_classes):
        index = np.where(target == i)
        diff = output[index] - target[index]
        corr = np.where(diff == 0)[0].shape[0] / 1.0
        total = index[0].shape[0] / 1.0
        print(corr, total)
        per_class_acc.append(corr / total)

    return per_class_acc


class ProgressMeter(object):
    """Default PyTorch pogress meter"""
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print('\t'.join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = '{:' + str(num_digits) + 'd}'
        return '[' + fmt + '/' + fmt.format(num_batches) + ']'


def adjust_learning_rate(optimizer, epoch, args):
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    lr = args.lr * (0.1 ** (epoch // 30))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


def get_iou(bb1, bb2):
    """
    Calculate the Intersection over Union (IoU) of two bounding boxes.

    Parameters
    ----------
    bb1 : dict
        Keys: {'x1', 'x2', 'y1', 'y2'}
        The (x1, y1) position is at the top left corner,
        the (x2, y2) position is at the bottom right corner
    bb2 : dict
        Keys: {'x1', 'x2', 'y1', 'y2'}
        The (x, y) position is at the top left corner,
        the (x2, y2) position is at the bottom right corner

    Returns
    -------
    float
        in [0, 1]
    """
    assert bb1['x1'] < bb1['x2']
    assert bb1['y1'] < bb1['y2']
    assert bb2['x1'] < bb2['x2']
    assert bb2['y1'] < bb2['y2']

    # determine the coordinates of the intersection rectangle
    x_left = max(bb1['x1'], bb2['x1'])
    y_top = max(bb1['y1'], bb2['y1'])
    x_right = min(bb1['x2'], bb2['x2'])
    y_bottom = min(bb1['y2'], bb2['y2'])

    if x_right < x_left or y_bottom < y_top:
        return 0.0

    # The intersection of two axis-aligned bounding boxes is always an
    # axis-aligned bounding box
    intersection_area = (x_right - x_left) * (y_bottom - y_top)

    # compute the area of both AABBs
    bb1_area = (bb1['x2'] - bb1['x1']) * (bb1['y2'] - bb1['y1'])
    bb2_area = (bb2['x2'] - bb2['x1']) * (bb2['y2'] - bb2['y1'])

    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    iou = intersection_area / float(bb1_area + bb2_area - intersection_area)
    assert iou >= 0.0
    assert iou <= 1.0
    return iou


def constant_init(module, val, bias=0):
    # Code adapted from https://github.com/open-mmlab/mmcv/blob/master/mmcv/cnn/utils/weight_init.py
    if hasattr(module, 'weight') and module.weight is not None:
        nn.init.constant_(module.weight, val)
    if hasattr(module, 'bias') and module.bias is not None:
        nn.init.constant_(module.bias, bias)


def normal_init(module, mean=0, std=1, bias=0):
    # Code adapted from https://github.com/open-mmlab/mmcv/blob/master/mmcv/cnn/utils/weight_init.py
    nn.init.normal_(module.weight, mean, std)
    if hasattr(module, 'bias') and module.bias is not None:
        nn.init.constant_(module.bias, bias)


@contextmanager
def _ignore_torch_cuda_oom():
    """
    A context which ignores CUDA OOM exception from pytorch.
    """
    try:
        yield
    except RuntimeError as e:
        # NOTE: the string may change?
        if "CUDA out of memory. " in str(e):
            pass
        else:
            raise


def retry_if_cuda_oom(func):
    """
    Makes a function retry itself after encountering
    pytorch's CUDA OOM error.
    It will first retry after calling `torch.cuda.empty_cache()`.

    If that still fails, it will then retry by trying to convert inputs to CPUs.
    In this case, it expects the function to dispatch to CPU implementation.
    The return values may become CPU tensors as well and it's user's
    responsibility to convert it back to CUDA tensor if needed.

    Args:
        func: a stateless callable that takes tensor-like objects as arguments

    Returns:
        a callable which retries `func` if OOM is encountered.

    Examples:
    ::
        output = retry_if_cuda_oom(some_torch_function)(input1, input2)
        # output may be on CPU even if inputs are on GPU

    Note:
        1. When converting inputs to CPU, it will only look at each argument and check
           if it has `.device` and `.to` for conversion. Nested structures of tensors
           are not supported.

        2. Since the function might be called more than once, it has to be
           stateless.
    """

    def maybe_to_cpu(x):
        try:
            like_gpu_tensor = x.device.type == "cuda" and hasattr(x, "to")
        except AttributeError:
            like_gpu_tensor = False
        if like_gpu_tensor:
            return x.to(device="cpu")
        else:
            return x

    @wraps(func)
    def wrapped(*args, **kwargs):
        with _ignore_torch_cuda_oom():
            return func(*args, **kwargs)

        # Clear cache and retry
        torch.cuda.empty_cache()
        with _ignore_torch_cuda_oom():
            return func(*args, **kwargs)

        # Try on CPU. This slows down the code significantly, therefore print a notice.
        logger = logging.getLogger(__name__)
        logger.info("Attempting to copy inputs of {} to CPU due to CUDA OOM".format(str(func)))
        new_args = (maybe_to_cpu(x) for x in args)
        new_kwargs = {k: maybe_to_cpu(v) for k, v in kwargs.items()}
        return func(*new_args, **new_kwargs)

    return wrapped



# coding: utf-8

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import print_function

import warnings
import logging

import torch
import torch.nn as nn

from gluoncv.torch.model_zoo import get_model, get_model_list
from gluoncv.torch.engine.config import get_cfg_defaults


def test_get_all_models():
    cfg = get_cfg_defaults()
    names = get_model_list()
    for name in names:
        if 'coot' in name: continue
        cfg.CONFIG.MODEL.NAME = name
        net = get_model(cfg)
        assert isinstance(net, nn.Module), '{}'.format(name)


def _test_model_list(model_list, use_cuda, x, pretrained, num_classes, **kwargs):
    cfg = get_cfg_defaults()
    for model in model_list:
        cfg.CONFIG.MODEL.NAME = model
        cfg.CONFIG.MODEL.PRETRAINED = pretrained
        cfg.CONFIG.DATA.NUM_CLASSES = num_classes
        net = get_model(cfg).eval()

        if use_cuda:
            net.cuda()
            x = x.cuda()
        with torch.no_grad():
            net(x)


def test_action_recognition_resnet_models():
    use_cuda = True
    if torch.cuda.device_count() == 0:
        use_cuda = False

    models = ['resnet18_v1b_kinetics400', 'resnet34_v1b_kinetics400', 'resnet50_v1b_kinetics400',
              'resnet101_v1b_kinetics400', 'resnet152_v1b_kinetics400']

    # single video frame with 224x224
    x = torch.rand(1, 3, 1, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)

    # multiple video frames with 224x224
    x = torch.rand(5, 3, 5, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)

    models = ['resnet50_v1b_sthsthv2']

    # single video frame with 224x224
    x = torch.rand(1, 3, 1, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=174)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=174)

    # multiple video frames with 224x224
    x = torch.rand(5, 3, 5, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=174)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=174)


def test_action_recognition_r2plus1d_models():
    use_cuda = True
    if torch.cuda.device_count() == 0:
        use_cuda = False

    models = ['r2plus1d_v1_resnet18_kinetics400', 'r2plus1d_v1_resnet34_kinetics400',
              'r2plus1d_v1_resnet50_kinetics400', 'r2plus1d_v2_resnet152_kinetics400',
              'r2plus1d_v1_resnet101_kinetics400', 'r2plus1d_v1_resnet152_kinetics400']

    x = torch.rand(2, 3, 16, 112, 112)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)


    models = ['r2plus1d_v1_resnet18_kinetics400', 'r2plus1d_v1_resnet34_kinetics400',
              'r2plus1d_v1_resnet50_kinetics400', 'r2plus1d_v2_resnet152_kinetics400']
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)


def test_action_recognition_i3d_models():
    use_cuda = True
    if torch.cuda.device_count() == 0:
        use_cuda = False

    models = ['i3d_resnet50_v1_kinetics400', 'i3d_resnet101_v1_kinetics400',
              'i3d_nl5_resnet50_v1_kinetics400', 'i3d_nl10_resnet50_v1_kinetics400',
              'i3d_nl5_resnet101_v1_kinetics400', 'i3d_nl10_resnet101_v1_kinetics400']

    x = torch.rand(2, 3, 32, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)

    models = ['i3d_resnet50_v1_sthsthv2']

    x = torch.rand(2, 3, 8, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=174)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=174)


def test_action_recognition_slowfast_models():
    use_cuda = True
    if torch.cuda.device_count() == 0:
        use_cuda = False

    models = ['slowfast_4x16_resnet50_kinetics400', 'slowfast_4x16_resnet101_kinetics400']
    x = torch.rand(2, 3, 32, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)

    models = ['slowfast_4x16_resnet50_kinetics400']
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)

    models = ['slowfast_8x8_resnet50_kinetics400', 'slowfast_8x8_resnet101_kinetics400']
    x = torch.rand(2, 3, 32, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)

    models = ['slowfast_16x8_resnet101_kinetics400', 'slowfast_16x8_resnet101_50_50_kinetics400']
    x = torch.rand(2, 3, 64, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)

    models = ['slowfast_16x8_resnet50_sthsthv2']
    x = torch.rand(2, 3, 64, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=174)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=174)


def test_action_recognition_i3d_slow_models():
    use_cuda = True
    if torch.cuda.device_count() == 0:
        use_cuda = False

    models = ['i3d_slow_resnet50_f32s2_kinetics400', 'i3d_slow_resnet101_f32s2_kinetics400']
    x = torch.rand(2, 3, 32, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)

    models = ['i3d_slow_resnet50_f8s8_kinetics400', 'i3d_slow_resnet101_f8s8_kinetics400']
    x = torch.rand(2, 3, 8, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)

    models = ['i3d_slow_resnet50_f16s4_kinetics400', 'i3d_slow_resnet101_f16s4_kinetics400']
    x = torch.rand(2, 3, 16, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)


def test_action_recognition_tpn_models():
    use_cuda = True
    if torch.cuda.device_count() == 0:
        use_cuda = False

    models = ['tpn_resnet50_f32s2_kinetics400', 'tpn_resnet101_f32s2_kinetics400']
    x = torch.rand(2, 3, 32, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)

    models = ['tpn_resnet50_f8s8_kinetics400', 'tpn_resnet101_f8s8_kinetics400']
    x = torch.rand(2, 3, 8, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)

    models = ['tpn_resnet50_f16s4_kinetics400', 'tpn_resnet101_f16s4_kinetics400']
    x = torch.rand(2, 3, 16, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)


def test_action_recognition_csn_models():
    use_cuda = True
    if torch.cuda.device_count() == 0:
        use_cuda = False

    models = ['ircsn_v2_resnet152_f32s2_kinetics400']
    x = torch.rand(2, 3, 32, 224, 224)
    _test_model_list(models, use_cuda, x, pretrained=False, num_classes=400)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=400)

def test_directpose_models():
    use_cuda = True
    if torch.cuda.device_count() == 0:
        use_cuda = False

    models = ['directpose_resnet50_lpf_fpn_coco']
    x = torch.rand(1, 3, 768, 1280)
    _test_model_list(models, use_cuda, x, pretrained=True, num_classes=1)


if __name__ == '__main__':
    import nose

    nose.runmodule()



from .boxes import Boxes, BoxMode, pairwise_iou
from .instances import Instances
from .masks import BitMasks, PolygonMasks, polygons_to_bitmask
from .image_list import ImageList
from .keypoints import Keypoints
from .beziers import Beziers



"""Directpose outputs"""
# pylint: disable=line-too-long, redefined-builtin, missing-class-docstring, unused-variable, consider-using-enumerate,unused-argument
import logging
import os
from typing import List
import numpy as np
import torch
from torch import nn
import torch.nn.functional as F
from torchvision.ops import nms

from ...data.structures import Instances, Boxes
from ...utils.comm import get_world_size
from ...nn.focal_loss import sigmoid_focal_loss_jit
from ...nn.smooth_l1_loss import smooth_l1_loss
from ...utils.comm import reduce_sum
from ...nn.nms import ml_nms, oks_nms, close_kpt_nms
from ...nn.iou_loss import IOULoss
from ...nn.keypoint_loss import WeightedMSELoss, HMFocalLoss

logger = logging.getLogger(__name__)

INF = 100000000

"""
Shape shorthand in this module:

    N: number of images in the minibatch
    L: number of feature maps per image on which RPN is run
    Hi, Wi: height and width of the i-th feature map
    4: size of the box parameterization

Naming convention:

    labels: refers to the ground-truth class of an position.

    reg_targets: refers to the 4-d (left, top, right, bottom) distances that parameterize the ground-truth box.

    logits_pred: predicted classification scores in [-inf, +inf];

    reg_pred: the predicted (left, top, right, bottom), corresponding to reg_targets

    ctrness_pred: predicted centerness scores

"""

def cat(tensors: List[torch.Tensor], dim: int = 0):
    """
    Efficient version of torch.cat that avoids a copy if there is only a single element in a list
    """
    assert isinstance(tensors, (list, tuple))
    if len(tensors) == 1:
        return tensors[0]
    return torch.cat(tensors, dim)


def compute_ctrness_targets(reg_targets):
    if len(reg_targets) == 0:
        return reg_targets.new_zeros(len(reg_targets))
    left_right = reg_targets[:, [0, 2]]
    top_bottom = reg_targets[:, [1, 3]]
    ctrness = (left_right.min(dim=-1)[0] / left_right.max(dim=-1)[0]) * \
              (top_bottom.min(dim=-1)[0] / top_bottom.max(dim=-1)[0])
    return torch.sqrt(ctrness)

class DirectPoseOutputs(nn.Module):
    def __init__(self, cfg):
        super(DirectPoseOutputs, self).__init__()

        self.focal_loss_alpha = cfg.CONFIG.MODEL.DIRECTPOSE.LOSS_ALPHA
        self.focal_loss_gamma = cfg.CONFIG.MODEL.DIRECTPOSE.LOSS_GAMMA
        self.center_sample = cfg.CONFIG.MODEL.DIRECTPOSE.CENTER_SAMPLE
        self.radius = cfg.CONFIG.MODEL.DIRECTPOSE.POS_RADIUS
        self.pre_nms_thresh_train = cfg.CONFIG.MODEL.DIRECTPOSE.INFERENCE_TH_TRAIN
        self.pre_nms_topk_train = cfg.CONFIG.MODEL.DIRECTPOSE.PRE_NMS_TOPK_TRAIN
        self.post_nms_topk_train = cfg.CONFIG.MODEL.DIRECTPOSE.POST_NMS_TOPK_TRAIN
        self.loc_loss_func = IOULoss(cfg.CONFIG.MODEL.DIRECTPOSE.LOC_LOSS_TYPE)
        self.kpt_l1_beta = cfg.CONFIG.MODEL.DIRECTPOSE.KPT_L1_beta
        self.enable_bbox_branch = cfg.CONFIG.MODEL.DIRECTPOSE.ENABLE_BBOX_BRANCH
        self.loss_on_locator = cfg.CONFIG.MODEL.DIRECTPOSE.LOSS_ON_LOCATOR
        self.enable_kpt_vis_branch = cfg.CONFIG.MODEL.DIRECTPOSE.KPT_VIS
        self.enable_close_kpt_nms = cfg.CONFIG.MODEL.DIRECTPOSE.CLOSEKPT_NMS
        # self.vis_res_dir = os.path.join('/workplace/data/motion_efs/home/manchenw/AdelaiDet/visualization',
        #                                 cfg.CONFIG.MODEL.WEIGHTS.split('/')[2])

        self.enable_hm_branch = cfg.CONFIG.MODEL.DIRECTPOSE.ENABLE_HM_BRANCH
        self.hm_type = cfg.CONFIG.MODEL.DIRECTPOSE.HM_TYPE
        self.predict_hm_offset = cfg.CONFIG.MODEL.DIRECTPOSE.HM_OFFSET
        self.hm_loss_type = cfg.CONFIG.MODEL.DIRECTPOSE.HM_LOSS_TYPE
        self.combine_hm_and_kpt = cfg.CONFIG.MODEL.DIRECTPOSE.REFINE_KPT
        self.hm_loss_weight = cfg.CONFIG.MODEL.DIRECTPOSE.HM_LOSS_WEIGHT
        self.g = self.get_gaussian_kernel(sigma=2)
        if self.hm_type == 'Gaussian':
            if self.hm_loss_type == 'mse':
                self.hm_bg_weight = cfg.CONFIG.MODEL.DIRECTPOSE.HM_MSELOSS_BG_WEIGHT
                self.hm_loss_weight = cfg.CONFIG.MODEL.DIRECTPOSE.HM_MSELOSS_WEIGHT
                self.hm_loss = WeightedMSELoss()
            elif self.hm_loss_type == 'focal':
                self.hm_bg_weight = 1.0
                hm_focal_alpha = cfg.CONFIG.MODEL.DIRECTPOSE.HM_FOCALLOSS_ALPHA
                hm_focal_beta = cfg.CONFIG.MODEL.DIRECTPOSE.HM_FOCALLOSS_BETA
                self.hm_loss = HMFocalLoss(hm_focal_alpha, hm_focal_beta)

        self.pre_nms_thresh_test = cfg.CONFIG.MODEL.DIRECTPOSE.INFERENCE_TH_TEST
        self.pre_nms_topk_test = cfg.CONFIG.MODEL.DIRECTPOSE.PRE_NMS_TOPK_TEST
        self.post_nms_topk_test = cfg.CONFIG.MODEL.DIRECTPOSE.POST_NMS_TOPK_TEST
        self.nms_thresh = cfg.CONFIG.MODEL.DIRECTPOSE.NMS_TH
        self.thresh_with_ctr = cfg.CONFIG.MODEL.DIRECTPOSE.THRESH_WITH_CTR

        self.num_classes = cfg.CONFIG.MODEL.DIRECTPOSE.NUM_CLASSES
        self.strides = cfg.CONFIG.MODEL.DIRECTPOSE.FPN_STRIDES
        self.num_kpts = cfg.CONFIG.MODEL.DIRECTPOSE.NUM_KPTS

        # generate sizes of interest
        soi = []
        prev_size = -1
        for s in cfg.CONFIG.MODEL.DIRECTPOSE.SIZES_OF_INTEREST:
            soi.append([prev_size, s])
            prev_size = s
        soi.append([prev_size, INF])
        self.sizes_of_interest = soi
        self.cnt = 0
        # TVM MODE
        self._tvm_mode = cfg.CONFIG.MODEL.TVM_MODE

    def _transpose(self, training_targets, num_loc_list):
        '''
        This function is used to transpose image first training targets to level first ones
        :return: level first training targets
        '''
        for im_i in range(len(training_targets)):
            training_targets[im_i] = torch.split(
                training_targets[im_i], num_loc_list, dim=0
            )

        targets_level_first = []
        for targets_per_level in zip(*training_targets):
            targets_level_first.append(
                torch.cat(targets_per_level, dim=0)
            )
        return targets_level_first

    def _get_ground_truth(self, locations, gt_instances, hm_size=None, images=None):
        num_loc_list = [len(loc) for loc in locations]

        # compute locations to size ranges
        loc_to_size_range = []
        for l, loc_per_level in enumerate(locations):
            loc_to_size_range_per_level = loc_per_level.new_tensor(self.sizes_of_interest[l])
            loc_to_size_range.append(
                loc_to_size_range_per_level[None].expand(num_loc_list[l], -1)
            )

        loc_to_size_range = torch.cat(loc_to_size_range, dim=0)
        locations = torch.cat(locations, dim=0)

        training_targets = self.compute_targets_for_locations(
            locations, gt_instances, loc_to_size_range, num_loc_list, images
        )

        training_targets["locations"] = [locations.clone() for _ in range(len(gt_instances))]
        training_targets["im_inds"] = [
            locations.new_ones(locations.size(0), dtype=torch.long) * i for i in range(len(gt_instances))
        ]

        # transpose im first training_targets to level first ones
        training_targets = {
            k: self._transpose(v, num_loc_list) for k, v in training_targets.items()
        }

        training_targets["fpn_levels"] = [
            loc.new_ones(len(loc), dtype=torch.long) * level
            for level, loc in enumerate(training_targets["locations"])
        ]

        # we normalize reg_targets by FPN's strides here
        bbox_reg_targets = training_targets["bbox_reg_targets"]
        kpt_reg_targets = training_targets["kpt_reg_targets"]
        for l in range(len(bbox_reg_targets)):
            bbox_reg_targets[l] = bbox_reg_targets[l] / float(self.strides[l])
            kpt_reg_targets[l][:, :, :2] = kpt_reg_targets[l][:, :, :2] / float(self.strides[l])

        if self.enable_hm_branch:
            if self.hm_type == 'Gaussian':
                training_targets["hm_targets"], training_targets["hm_weights"], training_targets["hm_offset"] = \
                    self.generate_gaussian_hm_target(gt_instances, hm_size, self.strides[0], sigma=2)
            elif self.hm_type == 'BinaryLabels':
                training_targets["hm_targets"], training_targets["hm_weights"], training_targets["hm_offset"] = \
                    self.generate_binary_hm_target(locations[0:num_loc_list[0]], gt_instances, hm_size)
            # visualize_hm(training_targets["hm_targets"], training_targets["hm_weights"], images)

        return training_targets

    def get_sample_region(self, boxes, strides, num_loc_list, loc_xs, loc_ys, bitmasks=None, radius=1):
        if bitmasks is not None:
            _, h, w = bitmasks.size()

            ys = torch.arange(0, h, dtype=torch.float32, device=bitmasks.device)
            xs = torch.arange(0, w, dtype=torch.float32, device=bitmasks.device)

            m00 = bitmasks.sum(dim=-1).sum(dim=-1).clamp(min=1e-6)
            m10 = (bitmasks * xs).sum(dim=-1).sum(dim=-1)
            m01 = (bitmasks * ys[:, None]).sum(dim=-1).sum(dim=-1)
            center_x = m10 / m00
            center_y = m01 / m00
        else:
            center_x = boxes[..., [0, 2]].sum(dim=-1) * 0.5
            center_y = boxes[..., [1, 3]].sum(dim=-1) * 0.5

        num_gts = boxes.shape[0]
        K = len(loc_xs)
        boxes = boxes[None].expand(K, num_gts, 4)
        center_x = center_x[None].expand(K, num_gts)
        center_y = center_y[None].expand(K, num_gts)
        # center_gt = boxes.new_zeros(boxes.shape)
        center_gts = []
        # no gt
        if center_x.numel() == 0 or center_x[..., 0].sum() == 0:
            return loc_xs.new_zeros(loc_xs.shape, dtype=torch.uint8)
        beg = 0
        for level, num_loc in enumerate(num_loc_list):
            end = beg + num_loc
            stride = strides[level] * radius
            xmin = center_x[beg:end] - stride
            ymin = center_y[beg:end] - stride
            xmax = center_x[beg:end] + stride
            ymax = center_y[beg:end] + stride
            # limit sample region in gt
            center_gts.append(torch.stack([
                torch.where(xmin > boxes[beg:end, :, 0], xmin, boxes[beg:end, :, 0]),
                torch.where(ymin > boxes[beg:end, :, 1], ymin, boxes[beg:end, :, 1]),
                torch.where(xmax > boxes[beg:end, :, 2], boxes[beg:end, :, 2], xmax),
                torch.where(ymax > boxes[beg:end, :, 3], boxes[beg:end, :, 3], ymax)], dim=2))
            # center_gt[beg:end, :, 0] = torch.where(xmin > boxes[beg:end, :, 0], xmin, boxes[beg:end, :, 0])
            # center_gt[beg:end, :, 1] = torch.where(ymin > boxes[beg:end, :, 1], ymin, boxes[beg:end, :, 1])
            # center_gt[beg:end, :, 2] = torch.where(xmax > boxes[beg:end, :, 2], boxes[beg:end, :, 2], xmax)
            # center_gt[beg:end, :, 3] = torch.where(ymax > boxes[beg:end, :, 3], boxes[beg:end, :, 3], ymax)
            beg = end
        center_gt = torch.cat(center_gts, dim=0)
        left = loc_xs[:, None] - center_gt[..., 0]
        right = center_gt[..., 2] - loc_xs[:, None]
        top = loc_ys[:, None] - center_gt[..., 1]
        bottom = center_gt[..., 3] - loc_ys[:, None]
        center_bbox = torch.stack((left, top, right, bottom), -1)
        inside_gt_bbox_mask = center_bbox.min(-1)[0] > 0
        return inside_gt_bbox_mask

    def compute_targets_for_locations(self, locations, targets, size_ranges, num_loc_list, images=None):
        labels = []
        bbox_reg_targets = []
        kpt_reg_targets = []
        target_inds = []
        xs, ys = locations[:, 0], locations[:, 1]

        num_targets = 0
        for im_i in range(len(targets)):
            targets_per_im = targets[im_i]
            bboxes = targets_per_im.gt_boxes.tensor
            labels_per_im = targets_per_im.gt_classes

            # no gt
            if bboxes.numel() == 0:
                labels.append(labels_per_im.new_zeros(locations.size(0)) + self.num_classes)
                bbox_reg_targets.append(locations.new_zeros((locations.size(0), 4)))
                kpt_reg_targets.append(locations.new_zeros((locations.size(0), self.num_kpts, 3)))
                target_inds.append(labels_per_im.new_zeros(locations.size(0)) - 1)
                continue

            area = targets_per_im.gt_boxes.area()
            kpts = targets_per_im.gt_keypoints.tensor

            l = xs[:, None] - bboxes[:, 0][None]
            t = ys[:, None] - bboxes[:, 1][None]
            r = bboxes[:, 2][None] - xs[:, None]
            b = bboxes[:, 3][None] - ys[:, None]
            reg_targets_per_im = torch.stack([l, t, r, b], dim=2)

            kpt_x_offset = kpts[:, :, 0][None] - xs[:, None, None]
            kpt_y_offset = kpts[:, :, 1][None] - ys[:, None, None]
            kpt_vis = kpts[:, :, 2][None].expand(len(locations), -1, -1)
            kpt_reg_targets_per_im = torch.stack([kpt_x_offset, kpt_y_offset, kpt_vis], dim=3)
            kpt_reg_targets_per_im[kpt_vis == 0] = 0

            if self.center_sample:
                if targets_per_im.has("gt_bitmasks_full"):
                    bitmasks = targets_per_im.gt_bitmasks_full
                else:
                    bitmasks = None
                is_in_boxes = self.get_sample_region(
                    bboxes, self.strides, num_loc_list, xs, ys,
                    bitmasks=bitmasks, radius=self.radius
                )
            else:
                is_in_boxes = reg_targets_per_im.min(dim=2)[0] > 0

            max_reg_targets_per_im = reg_targets_per_im.max(dim=2)[0]
            # limit the regression range for each location
            is_cared_in_the_level = \
                (max_reg_targets_per_im >= size_ranges[:, [0]]) & \
                (max_reg_targets_per_im <= size_ranges[:, [1]])

            locations_to_gt_area = area[None].repeat(len(locations), 1)
            locations_to_gt_area[is_in_boxes == 0] = INF
            locations_to_gt_area[is_cared_in_the_level == 0] = INF

            # if there are still more than one objects for a location,
            # we choose the one with minimal area
            locations_to_min_area, locations_to_gt_inds = locations_to_gt_area.min(dim=1)

            reg_targets_per_im = reg_targets_per_im[range(len(locations)), locations_to_gt_inds]
            kpt_reg_targets_per_im = kpt_reg_targets_per_im[range(len(locations)), locations_to_gt_inds]
            target_inds_per_im = locations_to_gt_inds + num_targets
            num_targets += len(bboxes)

            labels_per_im = labels_per_im[locations_to_gt_inds]
            labels_per_im[locations_to_min_area == INF] = self.num_classes

            labels.append(labels_per_im)
            bbox_reg_targets.append(reg_targets_per_im)
            kpt_reg_targets.append(kpt_reg_targets_per_im)
            target_inds.append(target_inds_per_im)

        return {
            "labels": labels,
            "bbox_reg_targets": bbox_reg_targets,
            "kpt_reg_targets": kpt_reg_targets,
            "target_inds": target_inds
        }

    def losses(self, logits_pred, bbox_reg_pred, kpt_reg_pred, kpts_locator_reg_pred, ctrness_pred, hms, hms_offset,
               locations, gt_instances, top_feats=None, images=None):
        """
        Return the losses from a set of FCOS predictions and their associated ground-truth.

        Returns:
            dict[loss name -> loss value]: A dict mapping from loss name to loss value.
        """

        if self.enable_hm_branch:
            training_targets = self._get_ground_truth(locations, gt_instances, hms.shape[2:], images)
        else:
            training_targets = self._get_ground_truth(locations, gt_instances)

        # Collect all logits and regression predictions over feature maps
        # and images to arrive at the same shape as the labels and targets
        # The final ordering is L, N, H, W from slowest to fastest axis.

        instances = Instances((0, 0))
        instances.labels = cat([
            # Reshape: (N, 1, Hi, Wi) -> (N*Hi*Wi,)
            x.reshape(-1) for x in training_targets["labels"]
        ], dim=0)
        instances.gt_inds = cat([
            # Reshape: (N, 1, Hi, Wi) -> (N*Hi*Wi,)
            x.reshape(-1) for x in training_targets["target_inds"]
        ], dim=0)
        instances.im_inds = cat([
            x.reshape(-1) for x in training_targets["im_inds"]
        ], dim=0)
        instances.bbox_reg_targets = cat([
            # Reshape: (N, Hi, Wi, 4) -> (N*Hi*Wi, 4)
            x.reshape(-1, 4) for x in training_targets["bbox_reg_targets"]
        ], dim=0, )
        instances.kpt_reg_targets = cat([
            # Reshape: (N, Hi, Wi, 17, 3) -> (N*Hi*Wi, 51)
            x.reshape(-1, self.num_kpts * 3) for x in training_targets["kpt_reg_targets"]
        ], dim=0, )
        instances.locations = cat([
            x.reshape(-1, 2) for x in training_targets["locations"]
        ], dim=0)
        instances.fpn_levels = cat([
            x.reshape(-1) for x in training_targets["fpn_levels"]
        ], dim=0)

        instances.logits_pred = cat([
            # Reshape: (N, C, Hi, Wi) -> (N, Hi, Wi, C) -> (N*Hi*Wi, C)
            x.permute(0, 2, 3, 1).reshape(-1, self.num_classes) for x in logits_pred
        ], dim=0, )

        output_kpt_dim = 3 if self.enable_kpt_vis_branch else 2
        instances.kpt_reg_pred = cat([
            # Reshape: (N, B, Hi, Wi) -> (N, Hi, Wi, B) -> (N*Hi*Wi, B)
            x.permute(0, 2, 3, 1).reshape(-1, self.num_kpts * output_kpt_dim) for x in kpt_reg_pred
        ], dim=0, )

        instances.ctrness_pred = cat([
            # Reshape: (N, 1, Hi, Wi) -> (N*Hi*Wi,)
            x.permute(0, 2, 3, 1).reshape(-1) for x in ctrness_pred
        ], dim=0, )

        if len(top_feats) > 0:
            instances.top_feats = cat([
                # Reshape: (N, -1, Hi, Wi) -> (N*Hi*Wi, -1)
                x.permute(0, 2, 3, 1).reshape(-1, x.size(1)) for x in top_feats
            ], dim=0, )
        if self.enable_bbox_branch:
            instances.bbox_reg_pred = cat([
                # Reshape: (N, B, Hi, Wi) -> (N, Hi, Wi, B) -> (N*Hi*Wi, B)
                x.permute(0, 2, 3, 1).reshape(-1, 4) for x in bbox_reg_pred
            ], dim=0, )
        if self.loss_on_locator:
            instances.kpt_locator_reg_pred = cat([
                # Reshape: (N, B, Hi, Wi) -> (N, Hi, Wi, B) -> (N*Hi*Wi, B)
                x.permute(0, 2, 3, 1).reshape(-1, self.num_kpts * 2) for x in kpts_locator_reg_pred
            ], dim=0, )

        if self.enable_hm_branch:
            return self.directpose_losses(instances, training_targets["hm_targets"], training_targets["hm_weights"],
                                          training_targets["hm_offset"], hms, hms_offset)
        else:
            return self.directpose_losses(instances)

    def directpose_losses(self, instances, hm_target=None, hm_weight=None, hm_offset_target=None, hm_pred=None, hm_offset_pred=None):
        num_classes = instances.logits_pred.size(1)
        assert num_classes == self.num_classes

        labels = instances.labels.flatten()

        pos_inds = torch.nonzero(labels != num_classes).squeeze(1)
        num_pos_local = pos_inds.numel()
        num_gpus = get_world_size()
        total_num_pos = reduce_sum(pos_inds.new_tensor([num_pos_local])).item()
        num_pos_avg = max(total_num_pos / num_gpus, 1.0)

        # prepare one_hot
        class_target = torch.zeros_like(instances.logits_pred)
        class_target[pos_inds, labels[pos_inds]] = 1

        class_loss = sigmoid_focal_loss_jit(
            instances.logits_pred,
            class_target,
            alpha=self.focal_loss_alpha,
            gamma=self.focal_loss_gamma,
            reduction="sum",
        ) / num_pos_avg

        instances = instances[pos_inds]
        instances.pos_inds = pos_inds

        ctrness_targets = compute_ctrness_targets(instances.bbox_reg_targets)
        ctrness_targets_sum = ctrness_targets.sum()
        loss_denorm = max(reduce_sum(ctrness_targets_sum).item() / num_gpus, 1e-6)
        instances.gt_ctrs = ctrness_targets

        if pos_inds.numel() > 0:
            ctrness_loss = F.binary_cross_entropy_with_logits(
                instances.ctrness_pred,
                ctrness_targets,
                reduction="sum"
            ) / num_pos_avg

            kpt_reg_targets_vis = instances.kpt_reg_targets.reshape(-1, self.num_kpts, 3)
            kpt_reg_vis = kpt_reg_targets_vis[:, :, 2] > 0
            pos_kpt = kpt_reg_vis.sum(dim=1)
            kpt_reg_pred = instances.kpt_reg_pred.reshape(len(instances), self.num_kpts, -1)[:, :, 0:2] * kpt_reg_vis[:, :, None]
            kpt_reg_loss = smooth_l1_loss(
                kpt_reg_pred,
                kpt_reg_targets_vis[:, :, 0:2],
                beta=self.kpt_l1_beta,
                reduction="none"
            )
            kpt_reg_loss = kpt_reg_loss.mean(dim=-1).sum(dim=1)
            kpt_reg_loss /= pos_kpt.clamp(min=1.0)
            kpt_reg_loss = (kpt_reg_loss * ctrness_targets * pos_kpt).sum() / (ctrness_targets * pos_kpt).sum().clamp(min=1.0)
            # kpt_reg_loss = kpt_reg_loss.sum() / loss_denorm

            if self.enable_bbox_branch:
                bbox_reg_loss = self.loc_loss_func(
                    instances.bbox_reg_pred,
                    instances.bbox_reg_targets,
                    ctrness_targets
                ) / loss_denorm

            if self.enable_kpt_vis_branch:
                kpt_vis_pred = instances.kpt_reg_pred.reshape(len(instances), self.num_kpts, -1)[:, :, 2]
                kpt_vis_target = kpt_reg_targets_vis[:, :, 2]
                kpt_vis_target[kpt_vis_target > 1] = 1
                total_kpt_pos = reduce_sum(kpt_vis_target.new_tensor([kpt_vis_target.sum()])).item()
                num_pos_kpt_avg = max(total_kpt_pos / num_gpus, 1.0)
                # kpt_vis_loss = sigmoid_focal_loss_jit(
                #     kpt_vis_pred,
                #     kpt_vis_target,
                #     alpha=self.focal_loss_alpha,
                #     gamma=self.focal_loss_gamma,
                #     reduction="sum",
                # ) / num_pos_kpt_avg
                kpt_vis_loss = F.binary_cross_entropy_with_logits(
                    kpt_vis_pred,
                    kpt_vis_target,
                    reduction="sum") / num_pos_kpt_avg

            if self.loss_on_locator:
                kpt_locator_reg_pred = instances.kpt_locator_reg_pred.reshape(-1, self.num_kpts, 2) * kpt_reg_vis[:, :, None]
                kpt_locator_reg_loss = smooth_l1_loss(
                    kpt_locator_reg_pred,
                    kpt_reg_targets_vis[:, :, 0:2],
                    beta=self.kpt_l1_beta,
                    reduction="none"
                )
                kpt_locator_reg_loss = kpt_locator_reg_loss.sum(dim=1).sum(dim=1)
                kpt_locator_reg_loss[pos_kpt > 0] /= pos_kpt[pos_kpt > 0]
                kpt_locator_reg_loss = kpt_locator_reg_loss.sum() / loss_denorm
        else:
            if self.enable_bbox_branch:
                bbox_reg_loss = instances.bbox_reg_pred.sum() * 0
            if self.enable_kpt_vis_branch:
                kpt_vis_loss = instances.kpt_reg_pred.sum() * 0
            if self.loss_on_locator:
                kpt_locator_reg_loss = instances.kpt_reg_pred.sum() * 0
            ctrness_loss = instances.ctrness_pred.sum() * 0
            kpt_reg_loss = instances.kpt_reg_pred.sum() * 0

        if self.enable_hm_branch:
            if self.hm_type == "Gaussian":
                if self.hm_loss_type == "mse":
                    hm_loss = self.hm_loss(hm_pred, hm_target, hm_weight) * self.hm_loss_weight
                elif self.hm_loss_type == "focal":
                    hm_loss = self.hm_loss(hm_pred, hm_target) * self.hm_loss_weight

                if self.predict_hm_offset:
                    hm_offset_target_vis = hm_offset_target.permute(0, 2, 3, 1, 4).reshape(-1, self.num_kpts, 3)
                    hm_offset_vis = hm_offset_target_vis[:, :, 2] > 0
                    hm_offset_pred = hm_offset_pred.permute(0, 2, 3, 1).reshape(-1, 2)[:, None, :].expand(-1, self.num_kpts, -1) \
                                     * hm_offset_vis[:, :, None]
                    pos_kpt = hm_offset_vis.sum(dim=1)
                    hm_offset_loss = smooth_l1_loss(
                        hm_offset_pred,
                        hm_offset_target_vis[:, :, 0:2],
                        beta=self.kpt_l1_beta,
                        reduction="none"
                    )
                    hm_offset_loss = hm_offset_loss.sum(dim=1).sum(dim=1)
                    hm_offset_loss[pos_kpt > 0] /= pos_kpt[pos_kpt > 0]
                    hm_offset_loss = hm_offset_loss.sum() / loss_denorm
            elif self.hm_type == "BinaryLabels":
                pos_joint_inds = torch.nonzero(hm_target.flatten()).squeeze(1)
                num_pos_local = pos_joint_inds.numel()
                total_num_pos = reduce_sum(pos_joint_inds.new_tensor([num_pos_local])).item()
                num_joint_pos_avg = max(total_num_pos / num_gpus, 1.0)
                hm_loss = sigmoid_focal_loss_jit(
                    hm_pred, hm_target, alpha=self.focal_loss_alpha,
                    gamma=self.focal_loss_gamma, reduction="sum") / max(num_joint_pos_avg, 1.0) * self.hm_loss_weight

        losses = {
            "loss_directpose_cls": class_loss,
            "loss_directpose_ctr": ctrness_loss,
            "loss_directpose_kpt_loc": kpt_reg_loss,
        }
        if self.enable_bbox_branch:
            losses["loss_directpose_bbox_loc"] = bbox_reg_loss
        if self.loss_on_locator:
            losses["loss_directpose_kpt_locator_loc"] = kpt_locator_reg_loss
        if self.enable_hm_branch:
            losses["loss_directpose_hm"] = hm_loss
            if self.predict_hm_offset:
                losses["loss_directpose_hm_offset"] = hm_offset_loss
        if self.enable_kpt_vis_branch:
            losses["loss_directpose_kpt_vis"] = kpt_vis_loss
        extras = {
            "instances": instances,
            "loss_denorm": loss_denorm,
        }
        return extras, losses

    def predict_proposals(
            self, logits_pred, bbox_reg_pred, kpt_reg_pred, ctrness_pred, hms, hms_offset,
            locations, image_sizes, top_feats=None, images=None
    ):
        if self.training:
            self.pre_nms_thresh = self.pre_nms_thresh_train
            self.pre_nms_topk = self.pre_nms_topk_train
            self.post_nms_topk = self.post_nms_topk_train
        else:
            self.pre_nms_thresh = self.pre_nms_thresh_test
            self.pre_nms_topk = self.pre_nms_topk_test
            self.post_nms_topk = self.post_nms_topk_test

        sampled_boxes = []

        bundle = {
            "l": locations, "o": logits_pred,
            "c": ctrness_pred,
            "kr": kpt_reg_pred, "s": self.strides,
        }
        if self.enable_bbox_branch:
            bundle["r"] = bbox_reg_pred
        if len(top_feats) > 0:
            bundle["t"] = top_feats

        if self._tvm_mode and not self.training:
            forward_for_single_feature_map_func = self.forward_for_single_feature_map_tvm
        else:
            forward_for_single_feature_map_func = self.forward_for_single_feature_map

        for i, per_bundle in enumerate(zip(*bundle.values())):
            # get per-level bundle
            per_bundle = dict(zip(bundle.keys(), per_bundle))
            # recall that during training, we normalize regression targets with FPN's stride.
            # we denormalize them here.
            l = per_bundle["l"]
            o = per_bundle["o"]
            r = per_bundle["r"] * per_bundle["s"] if "r" in bundle else None
            if self.enable_kpt_vis_branch:
                b, _, h, w = per_bundle["kr"].shape
                kr_reshape = per_bundle["kr"].reshape(b, self.num_kpts, 3, h, w)
                kr = kr_reshape[:, :, 0:2, :, :].reshape(b, -1, h, w) * per_bundle["s"]
                kr_vis = kr_reshape[:, :, 2, :, :]
            else:
                kr = per_bundle["kr"] * per_bundle["s"]
                kr_vis = None
            c = per_bundle["c"]
            t = per_bundle["t"] if "t" in bundle else None

            # DEBUG:
            # sampled_boxes.append((l, o, r, kr, kr_vis))
            # continue

            sampled_boxes.append(
                forward_for_single_feature_map_func(
                    l, o, r, kr, kr_vis, c, image_sizes, t
                )
            )

            for per_im_sampled_boxes in sampled_boxes[-1]:
                per_im_sampled_boxes.fpn_levels = l.new_ones(
                    len(per_im_sampled_boxes), dtype=torch.long
                ) * i

        boxlists = list(zip(*sampled_boxes))
        # ret = [Instances.cat(boxlist) for boxlist in boxlists]
        # output = [x.pred_boxes.tensor for x in ret] + [x.pred_keypoints for x in ret]
        # return output
        # boxlists = [Instances.cat(boxlist) for boxlist in boxlists]
        # boxlists = [self.select_over_all_level(Instances.cat(boxlist)) for boxlist in boxlists]
        boxlists = [Instances.cat(boxlist) for boxlist in boxlists]

        if self._tvm_mode and not self.training:
            pred_boxes = [boxlist.pred_boxes.tensor for boxlist in boxlists]
            pred_keypoints = [boxlist.pred_keypoints for boxlist in boxlists]
            pred_scores = [boxlist.scores for boxlist in boxlists]
            pred_ids = [boxlist.pred_classes for boxlist in boxlists]
            assert len(pred_ids) == 1, "tvm supports bs==1 only"
            if self.enable_hm_branch and self.combine_hm_and_kpt:
                pred_keypoints = self._refine_kpt(pred_boxes, pred_keypoints, hms, hms_offset, images, topk=40, thresh=0.1)
            # nms_input = torch.cat((, torch.arange(end=int(pred_boxes[0].shape[0])).to(pred_boxes[0].device).unsqueeze(-1)), dim=1)
            # print(nms_input.shape)
            nms_ret = nms(pred_boxes[0], pred_scores[0], self.nms_thresh)
            return pred_ids[0], nms_ret, pred_boxes[0], pred_scores[0], pred_keypoints[0]

        boxlists = self.select_over_all_levels(boxlists)
        if self.enable_hm_branch and self.combine_hm_and_kpt:
            boxlists = self.refine_kpt(boxlists, hms, hms_offset, images, topk=40, thresh=0.1)
        # visualize_kpt_offset(images, boxlists, vis_dir=self.vis_res_dir, cnt=self.cnt)
        # self.cnt += 1
        return boxlists

    def forward_for_single_feature_map_tvm(
            self, locations, logits_pred, bbox_reg_pred, kpt_reg_pred, kpt_vis_pred,
            ctrness_pred, image_sizes, top_feat=None
    ):
        N, C, H, W = logits_pred.shape

        # put in the same format as locations
        logits_pred = logits_pred.view(N, C, H, W).permute(0, 2, 3, 1)
        logits_pred = logits_pred.reshape(N, -1, C).sigmoid()
        kpt_regression = kpt_reg_pred.view(N, self.num_kpts * 2, H, W).permute(0, 2, 3, 1)
        kpt_regression = kpt_regression.reshape(N, -1, self.num_kpts, 2)
        ctrness_pred = ctrness_pred.view(N, 1, H, W).permute(0, 2, 3, 1)
        ctrness_pred = ctrness_pred.reshape(N, -1).sigmoid()
        if self.enable_bbox_branch:
            box_regression = bbox_reg_pred.view(N, 4, H, W).permute(0, 2, 3, 1)
            box_regression = box_regression.reshape(N, -1, 4)
        if self.enable_kpt_vis_branch:
            kpt_vis_pred = kpt_vis_pred.permute(0, 2, 3, 1).reshape(N, -1, self.num_kpts).sigmoid()
        if top_feat is not None:
            top_feat = top_feat.view(N, -1, H, W).permute(0, 2, 3, 1)
            top_feat = top_feat.reshape(N, H * W, -1)

        # if self.thresh_with_ctr is True, we multiply the classification
        # scores with centerness scores before applying the threshold.
        if self.thresh_with_ctr:
            logits_pred = logits_pred * ctrness_pred[:, :, None]
        # candidate_inds = logits_pred > self.pre_nms_thresh
        # pre_nms_top_n = candidate_inds.view(N, -1).sum(1)
        # pre_nms_top_n = pre_nms_top_n.clamp(max=self.pre_nms_topk)

        if not self.thresh_with_ctr:
            logits_pred = logits_pred * ctrness_pred[:, :, None]

        results = []
        for i in range(N):
            assert logits_pred.shape[-1] == 1, "single detection class only"
            per_box_cls = logits_pred[i]
            mask = (per_box_cls <= self.pre_nms_thresh)
            per_box_cls = per_box_cls.masked_fill(mask, 0)
            # per_box_cls = torch.where(mask, per_box_cls, per_box_cls * 0)
            per_class = per_box_cls * 0
            per_class = per_class.masked_fill(mask, -1)
            # per_candidate_inds = candidate_inds[i]
            # per_box_cls = per_box_cls[per_candidate_inds]

            # per_candidate_nonzeros = per_candidate_inds.nonzero()
            # per_box_loc = per_candidate_nonzeros[:, 0]
            # per_class = per_candidate_nonzeros[:, 1]

            per_kpt_regression = kpt_regression[i]
            per_kpt_regression = per_kpt_regression.masked_fill(mask.unsqueeze(-1), 0)
            # per_kpt_regression = per_kpt_regression[per_box_loc]
            # per_locations = locations[per_box_loc]
            per_locations = locations.clone()
            per_locations = per_locations.masked_fill(mask, 0)

            if self.enable_bbox_branch:
                per_box_regression = box_regression[i]
                per_box_regression = per_box_regression.masked_fill(mask, 0)
                # per_box_regression = per_box_regression[per_box_loc]
            if self.enable_kpt_vis_branch:
                per_kpt_vis = kpt_vis_pred[i]
                per_kpt_vis = per_kpt_vis.masked_fill(mask, 0)
                # per_kpt_vis = per_kpt_vis[per_box_loc]
            if top_feat is not None:
                per_top_feat = top_feat[i]
                per_top_feat = per_top_feat.masked_fill(mask, 0)
                # per_top_feat = per_top_feat[per_box_loc]

            # per_pre_nms_top_n = pre_nms_top_n[i]

            # It will only happen when there are more than 1000 person candidate in the image
            # if per_candidate_inds.sum().item() > per_pre_nms_top_n.item():
            #     per_box_cls, top_k_indices = per_box_cls.topk(per_pre_nms_top_n, sorted=False)
            #     per_class = per_class[top_k_indices]

            #     per_kpt_regression = per_kpt_regression[top_k_indices]
            #     per_locations = per_locations[top_k_indices]
            #     if self.enable_bbox_branch:
            #         per_box_regression = per_box_regression[top_k_indices]
            #     if self.enable_kpt_vis_branch:
            #         per_kpt_vis = per_kpt_vis[top_k_indices]
            #     if top_feat is not None:
            #         per_top_feat = per_top_feat[top_k_indices]

            if self.enable_kpt_vis_branch:
                keypoints = torch.stack([
                    per_locations[:, 0:1] + per_kpt_regression[:, :, 0],
                    per_locations[:, 1:] + per_kpt_regression[:, :, 1],
                    per_kpt_vis], dim=2)
            else:
                keypoints = torch.stack([
                    per_locations[:, 0:1] + per_kpt_regression[:, :, 0],
                    per_locations[:, 1:] + per_kpt_regression[:, :, 1],
                    torch.sqrt(per_box_cls)[:, None].expand(-1, self.num_kpts)], dim=2)

            if self.enable_bbox_branch:
                detections = torch.stack([
                    per_locations[:, 0] - per_box_regression[:, 0],
                    per_locations[:, 1] - per_box_regression[:, 1],
                    per_locations[:, 0] + per_box_regression[:, 2],
                    per_locations[:, 1] + per_box_regression[:, 3],
                ], dim=1)
            else:
                min_xy, _ = keypoints.min(dim=1)
                max_xy, _ = keypoints.max(dim=1)
                detections = torch.cat((min_xy[:, 0:2], max_xy[:, 0:2]), dim=1)
                detections = torch.stack((
                    detections[:, 0].clamp(min=0., max=float(image_sizes[i][1])),
                    detections[:, 1].clamp(min=0., max=float(image_sizes[i][0])),
                    detections[:, 2].clamp(min=0., max=float(image_sizes[i][1])),
                    detections[:, 3].clamp(min=0., max=float(image_sizes[i][0]))
                    ), dim=1)

            boxlist = Instances(image_sizes[i])
            boxlist.pred_boxes = Boxes(detections)
            boxlist.pred_keypoints = keypoints
            boxlist.scores = torch.sqrt(per_box_cls.squeeze(-1))
            boxlist.pred_classes = per_class.squeeze(-1)
            boxlist.locations = per_locations
            if top_feat is not None:
                boxlist.top_feat = per_top_feat
            results.append(boxlist)
            # print(boxlist.pred_boxes.tensor.shape, boxlist.pred_keypoints.shape, boxlist.scores.shape, boxlist.pred_classes.shape, boxlist.locations.shape)

        return results

    def forward_for_single_feature_map(
            self, locations, logits_pred, bbox_reg_pred, kpt_reg_pred, kpt_vis_pred,
            ctrness_pred, image_sizes, top_feat=None
    ):
        N, C, H, W = logits_pred.shape

        # put in the same format as locations
        logits_pred = logits_pred.view(N, C, H, W).permute(0, 2, 3, 1)
        logits_pred = logits_pred.reshape(N, -1, C).sigmoid()
        kpt_regression = kpt_reg_pred.view(N, self.num_kpts * 2, H, W).permute(0, 2, 3, 1)
        kpt_regression = kpt_regression.reshape(N, -1, self.num_kpts, 2)
        ctrness_pred = ctrness_pred.view(N, 1, H, W).permute(0, 2, 3, 1)
        ctrness_pred = ctrness_pred.reshape(N, -1).sigmoid()
        if self.enable_bbox_branch:
            box_regression = bbox_reg_pred.view(N, 4, H, W).permute(0, 2, 3, 1)
            box_regression = box_regression.reshape(N, -1, 4)
        if self.enable_kpt_vis_branch:
            kpt_vis_pred = kpt_vis_pred.permute(0, 2, 3, 1).reshape(N, -1, self.num_kpts).sigmoid()
        if top_feat is not None:
            top_feat = top_feat.view(N, -1, H, W).permute(0, 2, 3, 1)
            top_feat = top_feat.reshape(N, H * W, -1)

        # if self.thresh_with_ctr is True, we multiply the classification
        # scores with centerness scores before applying the threshold.
        if self.thresh_with_ctr:
            logits_pred = logits_pred * ctrness_pred[:, :, None]
        candidate_inds = logits_pred > self.pre_nms_thresh
        pre_nms_top_n = candidate_inds.view(N, -1).sum(1)
        pre_nms_top_n = pre_nms_top_n.clamp(max=self.pre_nms_topk)

        if not self.thresh_with_ctr:
            logits_pred = logits_pred * ctrness_pred[:, :, None]

        results = []
        for i in range(N):
            per_box_cls = logits_pred[i]
            per_candidate_inds = candidate_inds[i]
            per_box_cls = per_box_cls[per_candidate_inds]

            per_candidate_nonzeros = per_candidate_inds.nonzero(as_tuple=False)
            per_box_loc = per_candidate_nonzeros[:, 0]
            per_class = per_candidate_nonzeros[:, 1]

            per_kpt_regression = kpt_regression[i]
            per_kpt_regression = per_kpt_regression[per_box_loc]
            per_locations = locations[per_box_loc]

            if self.enable_bbox_branch:
                per_box_regression = box_regression[i]
                per_box_regression = per_box_regression[per_box_loc]
            if self.enable_kpt_vis_branch:
                per_kpt_vis = kpt_vis_pred[i]
                per_kpt_vis = per_kpt_vis[per_box_loc]
            if top_feat is not None:
                per_top_feat = top_feat[i]
                per_top_feat = per_top_feat[per_box_loc]

            per_pre_nms_top_n = pre_nms_top_n[i]

            # It will only happen when there are more than 1000 person candidate in the image
            if per_candidate_inds.sum().item() > per_pre_nms_top_n.item():
                per_box_cls, top_k_indices = per_box_cls.topk(per_pre_nms_top_n, sorted=False)
                per_class = per_class[top_k_indices]

                per_kpt_regression = per_kpt_regression[top_k_indices]
                per_locations = per_locations[top_k_indices]
                if self.enable_bbox_branch:
                    per_box_regression = per_box_regression[top_k_indices]
                if self.enable_kpt_vis_branch:
                    per_kpt_vis = per_kpt_vis[top_k_indices]
                if top_feat is not None:
                    per_top_feat = per_top_feat[top_k_indices]

            if self.enable_kpt_vis_branch:
                keypoints = torch.stack([
                    per_locations[:, 0:1] + per_kpt_regression[:, :, 0],
                    per_locations[:, 1:] + per_kpt_regression[:, :, 1],
                    per_kpt_vis], dim=2)
            else:
                keypoints = torch.stack([
                    per_locations[:, 0:1] + per_kpt_regression[:, :, 0],
                    per_locations[:, 1:] + per_kpt_regression[:, :, 1],
                    torch.sqrt(per_box_cls)[:, None].expand(-1, self.num_kpts)], dim=2)

            if self.enable_bbox_branch:
                detections = torch.stack([
                    per_locations[:, 0] - per_box_regression[:, 0],
                    per_locations[:, 1] - per_box_regression[:, 1],
                    per_locations[:, 0] + per_box_regression[:, 2],
                    per_locations[:, 1] + per_box_regression[:, 3],
                ], dim=1)
            else:
                if per_pre_nms_top_n == 0:
                    detections = torch.empty(0, 4).to(keypoints.device)
                else:
                    min_xy, _ = keypoints.min(dim=1)
                    max_xy, _ = keypoints.max(dim=1)
                    detections = torch.cat((min_xy[:, 0:2], max_xy[:, 0:2]), dim=1)
                    detections[:, 0] = detections[:, 0].clamp(min=0, max=image_sizes[i][1])
                    detections[:, 1] = detections[:, 1].clamp(min=0, max=image_sizes[i][0])
                    detections[:, 2] = detections[:, 2].clamp(min=0, max=image_sizes[i][1])
                    detections[:, 3] = detections[:, 3].clamp(min=0, max=image_sizes[i][0])

            boxlist = Instances(image_sizes[i])
            boxlist.pred_boxes = Boxes(detections)
            boxlist.pred_keypoints = keypoints
            boxlist.scores = torch.sqrt(per_box_cls)
            boxlist.pred_classes = per_class
            boxlist.locations = per_locations
            if top_feat is not None:
                boxlist.top_feat = per_top_feat
            results.append(boxlist)

        return results

    def _refine_kpt(self, pred_boxes, pred_keypoints, heatmaps, hms_offset=None, images=None, stride=8, topk=40, thresh=0.2):
        heatmaps = torch.clamp(heatmaps.sigmoid_(), min=1e-4, max=1-1e-4)
        # visualize_hm(heatmaps, None, images)
        heatmaps = _nms(heatmaps)
        hm_score, hm_inds, hm_ys, hm_xs = _topk_channel(heatmaps, K=topk)
        if hms_offset is not None:
            hp_offset = _transpose_and_gather_feat(hms_offset, hm_inds.view(len(pred_keypoints), -1))
            hp_offset = hp_offset.view(len(pred_keypoints), self.num_kpts, topk, 2)
            hm_xs = hm_xs + hp_offset[:, :, :, 0]
            hm_ys = hm_ys + hp_offset[:, :, :, 1]
        else:
            hm_xs += 0.5
            hm_ys += 0.5
        hm_xs *= stride
        hm_ys *= stride
        # visualize_hm_xy(hm_xs, hm_ys, hm_score, images, thresh=0.1)

        mask = (hm_score > thresh).float()
        hm_score = (1 - mask) * -1 + mask * hm_score
        hm_ys = (1 - mask) * (-10000) + mask * hm_ys
        hm_xs = (1 - mask) * (-10000) + mask * hm_xs
        hm_kps = torch.stack([hm_xs, hm_ys], dim=-1)

        for i in range(len(pred_keypoints)):
            bbox_pred = pred_boxes[i]
            kpt_regression = pred_keypoints[i][:, :, 0:2].permute(1, 0, 2).contiguous()
            kpt_regression_score = pred_keypoints[i][:, :, 2].permute(1, 0).contiguous()
            dist = (((kpt_regression[:, :, None, :] - hm_kps[i][:, None, :, :]) ** 2).sum(dim=3) ** 0.5)
            min_dist, min_ind = dist.min(dim=2)
            cur_hm_score = hm_score[i].gather(1, min_ind)
            cur_hm_kps = hm_kps[i].gather(1, min_ind[:, :, None].expand(-1, -1, 2))
            mask = (cur_hm_kps[:, :, 0] < bbox_pred[None, :, 0]) + (cur_hm_kps[:, :, 0] > bbox_pred[None, :, 2]) + \
                   (cur_hm_kps[:, :, 1] < bbox_pred[None, :, 1]) + (cur_hm_kps[:, :, 1] > bbox_pred[None, :, 3]) + \
                   (cur_hm_score < thresh) + \
                   (min_dist > (torch.max(bbox_pred[:, 2] - bbox_pred[:, 0], bbox_pred[:, 3] - bbox_pred[:, 1]) * 0.1)[None])
            mask = (mask > 0).float()
            kps_score = (1 - mask) * cur_hm_score + mask * kpt_regression_score
            mask = mask[:, :, None].expand(-1, -1, 2)
            kps = (1 - mask) * cur_hm_kps + mask * kpt_regression
            kps = torch.cat((kps, kps_score[:, :, None]), dim=2)
            pred_keypoints[i] = kps.permute(1, 0, 2).contiguous()
        return pred_keypoints

    def refine_kpt(self, boxlists, heatmaps, hms_offset=None, images=None, stride=8, topk=40, thresh=0.2):
        heatmaps = torch.clamp(heatmaps.sigmoid_(), min=1e-4, max=1-1e-4)
        # visualize_hm(heatmaps, None, images)
        heatmaps = _nms(heatmaps)
        hm_score, hm_inds, hm_ys, hm_xs = _topk_channel(heatmaps, K=topk)
        if hms_offset is not None:
            hp_offset = _transpose_and_gather_feat(hms_offset, hm_inds.view(len(boxlists), -1))
            hp_offset = hp_offset.view(len(boxlists), self.num_kpts, topk, 2)
            hm_xs = hm_xs + hp_offset[:, :, :, 0]
            hm_ys = hm_ys + hp_offset[:, :, :, 1]
        else:
            hm_xs += 0.5
            hm_ys += 0.5
        hm_xs *= stride
        hm_ys *= stride
        # visualize_hm_xy(hm_xs, hm_ys, hm_score, images, thresh=0.1)

        mask = (hm_score > thresh).float()
        hm_score = (1 - mask) * -1 + mask * hm_score
        hm_ys = (1 - mask) * (-10000) + mask * hm_ys
        hm_xs = (1 - mask) * (-10000) + mask * hm_xs
        hm_kps = torch.stack([hm_xs, hm_ys], dim=-1)

        for i in range(len(boxlists)):
            if len(boxlists[i]) == 0:
                continue
            bbox_pred = boxlists[i].pred_boxes.tensor
            kpt_regression = boxlists[i].pred_keypoints[:, :, 0:2].permute(1, 0, 2).contiguous()
            kpt_regression_score = boxlists[i].pred_keypoints[:, :, 2].permute(1, 0).contiguous()
            dist = (((kpt_regression[:, :, None, :] - hm_kps[i][:, None, :, :]) ** 2).sum(dim=3) ** 0.5)
            min_dist, min_ind = dist.min(dim=2)
            cur_hm_score = hm_score[i].gather(1, min_ind)
            cur_hm_kps = hm_kps[i].gather(1, min_ind[:, :, None].expand(-1, -1, 2))
            mask = (cur_hm_kps[:, :, 0] < bbox_pred[None, :, 0]) + (cur_hm_kps[:, :, 0] > bbox_pred[None, :, 2]) + \
                   (cur_hm_kps[:, :, 1] < bbox_pred[None, :, 1]) + (cur_hm_kps[:, :, 1] > bbox_pred[None, :, 3]) + \
                   (cur_hm_score < thresh) + \
                   (min_dist > (torch.max(bbox_pred[:, 2] - bbox_pred[:, 0], bbox_pred[:, 3] - bbox_pred[:, 1]) * 0.1)[None])
            mask = (mask > 0).float()
            kps_score = (1 - mask) * cur_hm_score + mask * kpt_regression_score
            mask = mask[:, :, None].expand(-1, -1, 2)
            kps = (1 - mask) * cur_hm_kps + mask * kpt_regression
            kps = torch.cat((kps, kps_score[:, :, None]), dim=2)
            boxlists[i].pred_keypoints = kps.permute(1, 0, 2).contiguous()
        return boxlists

    def select_over_all_levels(self, boxlists):
        num_images = len(boxlists)
        results = []
        for i in range(num_images):
            # multiclass nms
            result = ml_nms(boxlists[i], self.nms_thresh)
            # # Remove skeleton based NMS because seen from the result it is not helping
            # if self.enable_kpt_vis_branch:
            #     result = oks_nms(result, thresh=0.8, in_vis_thre=0.2)
            # if self.enable_close_kpt_nms:
            #     result = close_kpt_nms(result)
            number_of_detections = len(result)

            # Limit to max_per_image detections **over all classes**
            if number_of_detections > self.post_nms_topk > 0:
                cls_scores = result.scores
                image_thresh, _ = torch.kthvalue(
                    cls_scores.cpu(),
                    number_of_detections - self.post_nms_topk + 1
                )
                keep = cls_scores >= image_thresh.item()
                keep = torch.nonzero(keep).squeeze(1)
                result = result[keep]
            results.append(result)
        return results

    def select_over_all_level(self, boxlist):
        # multiclass nms
        result = ml_nms(boxlist, self.nms_thresh, fixed_size=False)
        # # Remove skeleton based NMS because seen from the result it is not helping
        # if self.enable_kpt_vis_branch:
        #     result = oks_nms(result, thresh=0.8, in_vis_thre=0.2)
        # if self.enable_close_kpt_nms:
        #     result = close_kpt_nms(result)
        number_of_detections = len(result)

        # Limit to max_per_image detections **over all classes**
        if number_of_detections > self.post_nms_topk > 0:
            cls_scores = result.scores
            image_thresh, _ = torch.kthvalue(
                cls_scores.cpu(),
                number_of_detections - self.post_nms_topk + 1
            )
            keep = cls_scores >= image_thresh.item()
            keep = torch.nonzero(keep).squeeze(1)
            result = result[keep]
        return result


    def get_gaussian_kernel(self, sigma):
        size = 6 * sigma + 3
        x = np.arange(0, size, 1, float)
        y = x[:, np.newaxis]
        x0, y0 = 3 * sigma + 1, 3 * sigma + 1
        g = np.exp(-((x - x0) ** 2 + (y - y0) ** 2) / (2 * sigma ** 2))
        return g

    def generate_gaussian_hm_target(self, gt_instances, hm_size, feat_stride=8, sigma=2):
        all_hms = []
        all_hms_weight = []
        all_hms_offset = []
        hm_h, hm_w = hm_size
        g = torch.from_numpy(self.g).float().to(gt_instances[0].gt_keypoints.device)

        for gt in gt_instances:
            hms = torch.zeros((self.num_kpts, hm_h, hm_w), dtype=torch.float32, device=gt.gt_keypoints.device)
            if self.hm_loss_type == 'mse':
                hms_weight = 2 * torch.ones((self.num_kpts, hm_h, hm_w), dtype=torch.float32, device=gt.gt_keypoints.device)
            if self.predict_hm_offset:
                hms_offset = torch.zeros((self.num_kpts, hm_h, hm_w, 3), dtype=torch.float32, device=gt.gt_keypoints.device)

            joints = gt.gt_keypoints.tensor
            for p in joints:
                for idx, pt in enumerate(p):
                    if pt[2] > 0:
                        x = int(pt[0] / feat_stride + 0.5)
                        y = int(pt[1] / feat_stride + 0.5)
                        if x < 0 or y < 0 or x >= hm_w or y >= hm_h:
                            continue

                        ul = int(np.floor(x - 3 * sigma - 1)), int(np.floor(y - 3 * sigma - 1))
                        br = int(np.floor(x + 3 * sigma + 2)), int(np.floor(y + 3 * sigma + 2))

                        c, d = max(0, -ul[0]), min(br[0], hm_w) - ul[0]
                        a, b = max(0, -ul[1]), min(br[1], hm_h) - ul[1]

                        cc, dd = max(0, ul[0]), min(br[0], hm_w)
                        aa, bb = max(0, ul[1]), min(br[1], hm_h)

                        hms[idx, aa:bb, cc:dd] = torch.max(hms[idx, aa:bb, cc:dd], g[a:b, c:d])
                        if self.hm_loss_type == 'mse':
                            hms_weight[idx, aa:bb, cc:dd] = 1.0
                        if self.predict_hm_offset:
                            hms_offset[idx, y, x, 0] = pt[0] / feat_stride - x
                            hms_offset[idx, y, x, 1] = pt[1] / feat_stride - y
                            hms_offset[idx, y, x, 2] = 1.0

            all_hms.append(hms)
            if self.hm_loss_type == 'mse':
                hms_weight[hms_weight == 2] = self.hm_bg_weight
                all_hms_weight.append(hms_weight)
            if self.predict_hm_offset:
                all_hms_offset.append(hms_offset)

        all_hms = torch.stack(all_hms, dim=0)
        if self.hm_loss_type == 'mse':
            all_hms_weight = torch.stack(all_hms_weight, dim=0)
        if self.predict_hm_offset:
            all_hms_offset = torch.stack(all_hms_offset, dim=0)

        return all_hms, all_hms_weight, all_hms_offset

    def generate_binary_hm_target(self, locations, gt_instances, hm_size):
        all_hms = []
        locations_x = locations[:, 0:1]
        locations_y = locations[:, 1:2]
        for gt in gt_instances:
            target = torch.zeros((self.num_kpts, hm_size[0] * hm_size[1]), dtype=torch.float32, device=gt.gt_keypoints.device)

            kps = gt.gt_keypoints.tensor.reshape(-1, 3)
            kps_types = torch.arange(kps.size(0), dtype=torch.int, device=kps.device) % self.num_kpts
            if (kps[:, 2] > 0).sum() > 0:
                visible_kps = kps[kps[:, 2] > 0]
                visible_kps_types = kps_types[kps[:, 2] > 0]

                distances_x = torch.abs(locations_x - visible_kps[:, 0][None])
                distances_y = torch.abs(locations_y - visible_kps[:, 1][None])
                distances = torch.max(distances_x, distances_y)
                _, min_inds = distances.min(dim=0)

                for (kpt_idx, heatmap_loc) in zip(visible_kps_types, min_inds):
                    target[kpt_idx, heatmap_loc] = 1
            all_hms.append(target.reshape(self.num_kpts, hm_size[0], hm_size[1]))
        all_hms = torch.stack(all_hms, dim=0)
        return all_hms, None, None

def _nms(heat, kernel=3):
    pad = (kernel - 1) // 2

    hmax = nn.functional.max_pool2d(
        heat, (kernel, kernel), stride=1, padding=pad)
    keep = (hmax == heat).float()
    return heat * keep

def _topk_channel(scores, K=40):
    batch, cate, height, width = scores.size()

    topk_scores, topk_inds = torch.topk(scores.view(batch, cate, -1), K)

    topk_inds = topk_inds % (height * width)
    topk_ys = (topk_inds / width).int().float()
    topk_xs = (topk_inds % width).int().float()

    return topk_scores, topk_inds, topk_ys, topk_xs

def _gather_feat(feat, ind):
    dim = feat.size(2)
    ind = ind.unsqueeze(2).expand(ind.size(0), ind.size(1), dim)
    feat = feat.gather(1, ind)
    return feat

def _transpose_and_gather_feat(feat, ind):
    feat = feat.permute(0, 2, 3, 1).contiguous()
    feat = feat.view(feat.size(0), -1, feat.size(3))
    feat = _gather_feat(feat, ind)
    return feat
